from typing import Union, Tuple
from pathlib import Path
from mkdocs.structure.pages import Page



Ref = SnipDescription = str

UriPath = str
""" String representing an uri path. May be rooted at cwd or docs_dir, depending on the
    "logic" it refers to.
"""

UriDocsPathStr = str
""" String representing an uri path, rooted at the docs_dir """

UriCwdPathStr = str
""" String representing an uri path, rooted at the cwd """

SourceUsedRef = Union[Page, UriDocsPathStr]
""" Union[ Page, UriDocsPath ] """

#----------------------

PathCwd = Path
""" Path instance, rooted at the cwd """

PathDocs = Path
""" Path instance, rooted at the docs_dir """

SourceRef = Union[ Page, PathCwd ]
""" Union[ Page, PathCwd ] """





from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Set

from collections import defaultdict
from pathlib import Path


from mkdocs_addresses.exceptions import DuplicateIdentifierError
from mkdocs_addresses.static_handler.page_data_tracker import PageDataTracker
from .types_aliases import Ref, PathCwd, SnipDescription, SourceRef







SRC_TO_IDS_TYPE = set




class References(PageDataTracker):
    """
    Keep track of each reference definition (headers, explicit ids, files, external links in a
    dedicated file):
        - identifier
        - associated source to point to when the identifier will be used
        - file location of the identifier (as "source uri")
          WARNING: the path is relative to the cwd.
        - code snippets associated to each identifier, if use_vsc is True
        - ensure that all identifiers (references or those used for the code snippets) are
          unique, otherwise raise DuplicateIdentifierError

    Note about the external links: they are stored only to build the code snippets/info file
    about the available references? Mkdocs actually already handles their transformation into
    proper urls.
    """

    id_to_data:  Dict[Ref,Any]
    """ Dict "identifier -> data" in the built site. Depending on what's registered:

        | Identifier | Data |
        |:-|:-|
        | A link/img ref<br>(`--link`/`!!img`) | Url page/file in the rendered website, relative to the docs_dir |
        | A file ref (`++file`) | Url of the file in the rendered website, relative to the docs_dir |
        | An external link ref (`BusService`) | the url of the link |
        | Description of a snippet | complete code, as string, of the item for the snippet |
    """

    other_entries:  List[Any] = field(default_factory=list)
    """ List of any code snippet the user might have added in the target file (when using vsc) """

    snippets: Dict[SnipDescription,str] = field(default_factory=dict)
    """ Dict "snippet description -> full snippet item code (as string)"
        The snippet description is the key used to identify a snippet in the final file.
    """

    _old_refs: Set[Ref]
    """ Registered references during the previous build """


    UNIQUE_SRC = True       # override


    def __init__(self, uni_docs_dir:PathCwd):
        super().__init__(uni_docs_dir)
        self.id_to_data     = {}
        self.snippets       = {}
        self.other_entries  = []
        self._old_refs      = set()


    def __str__(self):
        return f"{ self.__class__.__name__ }({ self.id_to_data })"


    #--------------------------------------------------------------
    # Snippets related logistic


    def clear_other_snippets(self):
        self.other_entries.clear()

    def store_other_snippets(self, snippets:Iterable[str]):
        self.other_entries.extend(snippets)


    def add_snippet(self, src:SourceRef, identifier:str, snippet:str):
        here = self.format_source(src)
        if identifier in self.snippets:
            there = next(iter(self.id_to_sources[identifier]))
            raise DuplicateIdentifierError(
                f"{ identifier !r} is already used.\n"
                f"  - Attempt to add it to snippets\n"
                f"  - Adding from file { here }\n"
                f"  - Already registered with file {there}\n"
            )

        self.snippets[identifier] = snippet
        self.id_to_sources[identifier].add(here)
        self.source_to_ids[here].add(identifier)



    def build_json_snippets(self):
        snippets = sorted(( *self.other_entries,  *self.snippets.values() ))
        code = "{\n" + ",\n".join(snippets) + "\n}"
        return code



    #--------------------------------------------------------------


    def add_id(
        self,
        source:SourceRef,
        identifier:str,
        data:Any,
        id_declared_as:str=None,
):
        """ Store the given data, associated to the given id, for the given page.
            Raise DuplicateIdentifierError if @id has already been used before.
            @source: uri if the source file. This will be converted to an OsCwdPath.
            @id_declared_as: original version of the identifier in the source file, if
                             different from id.
        """
        super().add_id(source, identifier, id_declared_as)
        self.id_to_data[identifier] = data


    def remove_source(self, source:SourceRef):
        src,ids = super().remove_source(source)
        for identifier in ids:
            self.id_to_data.pop(identifier, None)
            self.snippets.pop(identifier, None)
        return src, ids                         # just to respect the interface of the parent...
                                                # but never used (so far)

    def remove_id(self, identifier:str):
        self.id_to_data.pop(identifier, None)
        self.snippets.pop(identifier, None)
        srcs = super().remove_id(identifier)
        return srcs                             # just to respect the interface of the parent...
                                                # but never used (so far)


    #---------------------------------


    def gen_refs_items(self):
        return iter(self.id_to_data.items())


    def get_ref_target_address(self, identifier:str):
        """ Return the data associated to an identifier """
        return self.id_to_data[identifier]


    #---------------------------------


    def archive_current(self, uni_docs_dir:PathCwd):
        super().archive_current(uni_docs_dir)
        self._old_refs = self.id_to_data.copy()



    def get_update_refs_definitions_info(self):
        """ Gather intel about all the references definitions, comparing the current state
            with the previous run of the plugin, and return:
                - a set of all the references that have been suppressed
                - a set of all the references that have been added

            This is mostly to give feedback to the user in the console.
        """
        fresh      = set(self.id_to_data)
        old        = set(self._old_refs)
        suppressed = old-fresh
        added      = fresh-old
        return (
            { ref: self._old_refs[ref]  for ref in suppressed },
            { ref: self.id_to_data[ref] for ref in added },
        )



import os
import re

from typing import Tuple
from enum import auto, IntEnum
from textwrap import dedent
from dataclasses import dataclass

from .config_plugin import PLUGIN_MARK


OPEN, CLOSE = '{', '}'



class RefKind(IntEnum):
    Link = auto()
    File = auto()
    Img  = auto()
    Ext  = auto()
    Include = auto()





@dataclass
class VscAutoCompletion:
    """
    A VscAutoCompletion instance holds the whole logic to build one code snippets
    for a given kind of data
    """

    kind: RefKind               # Snippet first char (case sensitive)
    head: str                   # Prefix identifier (used to spot them in the html code)


    @staticmethod
    def setup_class():
        VscAutoCompletion.__LINK_BUILDER = {
            formatter.kind: formatter for formatter in [
                VscAutoAddress( RefKind.Link,    '--'),                     # internal links
                VscAutoAddress( RefKind.File,    '++'),                     # any internal file
                VscAutoAddress( RefKind.Img,     '!!', snippet_start='!'),  # images
                VscAutoAddress( RefKind.Ext,      '', tail_wrap=('[',']'), attrs="target=_blank"),  # External links
                VscAutoInclude( RefKind.Include, '::')
        ]}
        VscAutoCompletion.__POTENTIAL_REF_PATTERN = re.compile(
            '|'.join( re.escape(o.head) for o in VscAutoCompletion.__LINK_BUILDER.values() if o.head)
        )

    @classmethod
    def href_is_possible_ref(cls, href):
        return cls.__POTENTIAL_REF_PATTERN.match(href)

    @classmethod
    def get_completer_for(cls, kind:RefKind) -> 'VscAutoCompletion':
        return cls.__LINK_BUILDER[kind]



    def get_final_identifier(self, identifier:str, with_head=True):
        """ Add the "head_link" part to the identifier if needed, and format the identifier
            to avoid they generate warnings during the builds
        """
        return with_head * self.head + identifier


    def build_snippet(self, identifier, src_link, **_):
        """ Generate code snippet entries for the given (bare) identifier, pointing toward the
            given source (which will, in the end, be the targeted element).
            @identifier: the bare identifier, without the "head" part ("!!", "--", ...)
            @src_link is the path of the source, relative to the cwd ("development wise")
            Automatically add the "shorthand" version of the snippet, if the kind requires it.
        """
        raise NotImplementedError("Subclasses must override this method")


    def _build_snippet(self, prefix:str, body:str, src_link:str, name_prefix=''):
        if name_prefix:
            name_prefix += " - "
        json_id = f"{ name_prefix }{ self.kind.name }: { src_link !r}"
        return json_id, dedent(f"""\
            "{ json_id }": { OPEN }
                "prefix": "{ prefix }",
                "scope": "markdown",
                "body": ["{ body }"],
                "description": "{ PLUGIN_MARK }"
            { CLOSE }""")




@dataclass
class VscAutoAddress(VscAutoCompletion):
    """ Build snippets for links, ids, extras, files, images """

    snippet_start: str = ''     # Element coming before the "[...]" initial part
    tail_wrap: Tuple[str,str] = '(',')'     # open+close for the last section (after "[...]")
    with_short: bool = True     # Also generate the short snippet version (head+identifier only)
    attrs: str = ""             # string of attributes to always add at the end of the body


    def get_final_identifier(self, identifier:str, with_head=True):

        assets = 'assets/'
        if identifier.startswith(assets):
            identifier = identifier[len(assets):]
        return with_head * self.head + identifier.replace('.','_')


    def build_snippet(self, identifier, src_link, **_):

        clean_id = self.get_final_identifier(identifier, False)
        prefix   = f"{ self.kind.name }.{ clean_id }"       # Img.identifier or so...
        body     = self.__build_body(clean_id)
        yield self._build_snippet(prefix, body, src_link, "Md")

        if self.with_short:
            prefix = self.head + clean_id                   # !!identfier or so...
            body   = f"${ OPEN }0:{ self.head }{ CLOSE }{ clean_id }"
            yield self._build_snippet(prefix, body, src_link, "Reference")


    def __build_body(self, clean_id:str):
        """ helper... """
        L,R     = self.tail_wrap
        content = self._get_content(clean_id)
        body    = self.snippet_start + "[${0:" + content + "}]" + f"{ L }{ self.head }{ clean_id }{ R }"
        if self.attrs:
            body += "{: " + self.attrs + " }"
        return body

    def _get_content(self, clean_id:str):
        if self.kind == RefKind.Ext:
            return clean_id
        return 'content'






@dataclass
class VscAutoInclude(VscAutoCompletion):
    """ Build snippets for markdown code inclusions (ie: "--8<--" ) """


    # pylint: disable-next=arguments-differ
    def build_snippet(self, identifier:str, src_link:str, *, root_inclusion:str):
        """ root_inclusion: one of the path s given in plugin.inclusions (as uri) """

        _, ext    = os.path.splitext(identifier)
        short_uri = identifier[ len(root_inclusion)+1 : ]       # note
        body      = f'--8<-- \\"{ identifier }\\"'
        prefix    = f"{ self.head }{ ext.strip('.') } { short_uri }"
        snippet   = self._build_snippet(prefix, body, identifier)
        yield snippet




VscAutoCompletion.setup_class()         # Trigger global definitions in the top level class



import json
import re

from typing import DefaultDict, List, Literal, Set, NamedTuple, Union
from collections import Counter, defaultdict
from operator import itemgetter
from pathlib import Path

from bs4.element import Tag

from mkdocs.structure.pages import Page
from mkdocs.structure.files import Files, File
from mkdocs.structure.nav import Navigation

from mkdocs.plugins import BasePlugin, event_priority
from mkdocs.config.defaults import MkDocsConfig
from mkdocs_addresses import path_manager

from mkdocs_addresses.soup_excluding_codes import PluginSoupExcludingCodes
from mkdocs_addresses.static_handler.types_aliases import PathCwd, SourceRef, UriDocsPathStr


from .exceptions import (
    OutdatedReferenceError,
    UnknownExternalLinkIdentifierError,
    UnknownIdentifierError,
    DumpOnlyError,
    InvalidIdentifierError,
    LeadingSlashAddressError,
    NonAbsoluteAddressError,
)
from .config_plugin import (
    AddressAddressesConfig,
    ConfigDescriptor,
    NO_DUMP,
    DUMP_ONLY,
    PLUGIN_MARK,
)
from .static_handler import StaticHandler
from .toolbox import build_other_snippet, extract_external_links_refs_in_md, plugin_dump_padder, fatal_if
from .vsc_auto_completion_handler import RefKind, VscAutoCompletion
from .addresses_checker import AddressChecker, AddressKind
from .logger import LogMessages, logger, addresses_auto_log_plugin_on





#from ._test_data_extraction import LiveDataExtractor               ##>
#extractor = LiveDataExtractor()
#logger.error('Creating global LiveDataExtractor object')           ##<





class SoupTag(NamedTuple):
    tag: Tag
    attr: str
    kind: RefKind
    identifier: str








#-----------------------------------------------------------------------------
# Functions used for decorated methods (StaticHandler/cache related logistic):
#
# These functions are here to entirely decorelate the cache logic from the
# actual signatures of the plugin methods.
#-----------------------------------------------------------------------------



def plugin_file_prop_extractor(prop:str):
    def getter(self:'AddressAddressesPlugin', *_, **__) -> PathCwd:
        """ Extract a filename from a plugin property.
            WARNING: those are relative to cwd, not to docs_dir!
        """
        file: PathCwd = getattr(self, prop)
        return file
    return getter


def cwd_path_arg_extractor_page_hook(plugin:'AddressAddressesPlugin', *_, page:Page, **_kw) -> PathCwd:
    """ Extract the PathCwd of the page argument from a page related plugin event """
    return path_manager.get_cwd_path_from_docs_dir_and_uri(plugin.uni_docs_dir, page.file.src_uri)


def page_arg_extractor_page_hooks(_plugin, *_args, page:Page, **_kw) -> Page:
    """ Extract the page argument for the StaticHandler """
    return page


def cwd_path_arg_extractor_other_files(_plugin, path:PathCwd, _file:File) -> PathCwd:
    """ Extract the PathCwd argument for the StaticHandler """
    return path


def cwd_path_arg_extractor_inclusions(_plugin, file:str, _inclusion_path:str) -> PathCwd:
    """ Extract the PathCwd to the inclusions file from the arguments, for the StaticHandler """
    return path_manager.to_os(file)


#----------------------------------------


def never_skip(*_,**__):
    return False

def skip_if_no_vsc_or_no_snippets_file_or_wrong_extension(plugin:'AddressAddressesPlugin', *_, **__) -> bool:
    no_vsc    = not plugin.use_vsc
    no_file   = not plugin.dump_snippets_file.is_file()
    wrong_ext = not plugin.dump_snippets_file.suffix == '.code-snippets'
    return no_vsc or no_file or wrong_ext

def skip_if_no_external_links(plugin:'AddressAddressesPlugin', *_, **__) -> bool:
    return not plugin.external_links_file

def skip_if_only_dump(plugin:'AddressAddressesPlugin', *_, **__) -> bool:
    """ Check special reasons to not apply the on_page_context event """
    return plugin.dump_action == DUMP_ONLY

def skip_if_file_already_registered(plugin:'AddressAddressesPlugin', cwd_path:PathCwd, _:File) -> bool:
    return plugin.global_handler.references.has_source(cwd_path)


#-------------------------------------------------------------------------------










class AddressAddressesPlugin(BasePlugin[AddressAddressesConfig]):

    activate_cache: bool = ConfigDescriptor()

    docs_dir: str = ConfigDescriptor()
    """ uri segment, from cwd to the directory containing the documentation files """

    dump_action: str = ConfigDescriptor()
    """ 'normal', 'none', 'only' """

    dump_snippets_file: PathCwd = ConfigDescriptor()
    """ Relative to cwd """

    external_links_file: Union[PathCwd, Literal[""]] = ConfigDescriptor()
    """ Relative to cwd """

    fail_fast: bool = ConfigDescriptor()
    """ If false, information about the exceptions raised by the plugin are logged to the console
        but are caught, to let the executions go.
    """

    imgs_extension_pattern: re.Pattern = ConfigDescriptor()
    """ Regex pattern used to decide what should be considered a file or an image.
        This will decide the prefix used ('!!' for images and '++' for files)
    """

    ignore_auto_headers: bool = ConfigDescriptor()
    """ If true, the plugin will try to spot headers without specific id attribute and won't
        generate identifier for them (this allow to have the same title in different places
        without raising a DuplicateIdentifierError).
    """

    ignored_classes: Set[str] = ConfigDescriptor()
    """ When gathering references in the html code, tags holding one of these classes
        will be ignored
    """

    ignored_ids: Set[str] = ConfigDescriptor()
    """ When gathering references in the html code, tags holding one of these ids
        will be ignored
    """

    inclusions: List[PathCwd] = ConfigDescriptor()
    """ Relative to the cwd. List of the directories containing files that can be included
        using --8<-- in the docs.
    """

    links_white_list: Set[str] = ConfigDescriptor()
    """ References or actual addresses referenced here will be ignored, both when searching
        for references, and to decide if an address must be checked or not.
    """

    links_white_list_pattern: re.Pattern = ConfigDescriptor()
    """ Same as links_white_list, but using a regex pattern. """

    more_verbose: bool = ConfigDescriptor()
    """ If true, the plugin will log more thing at INFO level. This allow to get more feedback
        without using "mkdocs serve -v', which will literally get the console  swarmed.
    """

    plugin_off: bool = ConfigDescriptor()
    """ Bypass all behaviors if True """

    uni_docs_dir: PathCwd = ConfigDescriptor()
    """ Relative to cwd """

    use_directory_urls: bool = ConfigDescriptor()
    """ keep track of the config value """

    use_vsc: bool = ConfigDescriptor()
    # Note: if this needs to evolve further, to offer support for autocompletion with other softwares,
    #       this should probably become "use_auto_completion" (to check it matches every use case)

    verify_only: Set[UriDocsPathStr] = ConfigDescriptor()
    """ Set of files to verify during on_page_context. If empty, all (updated) files are checked,
        otherwise, only the files listed here are checked (if they have been updated, when the
        cache is activated).
    """




    # Plugin specific:

    global_handler: StaticHandler
    exceptions_counter: DefaultDict[str,List[str]] = None



    def __str__(self):
        plugin = {
            prop: list(val) if isinstance((val:=getattr(self,prop)),set) else
                  val.pattern if isinstance(val, re.Pattern) else
                  ['...'] if prop=='global_handler' else
                  str(val) if isinstance(val, Path) else
                  val
            for prop in sorted(self.__class__.__annotations__.keys())
        }
        plugin.update({
            "references": dict(self.global_handler.references.id_to_data),
        })
        return f'{ self.__class__.__name__ }:\n{ json.dumps(plugin, indent=4) }'




    #-------------------------------------------------------------------
    #                           Plugin Events
    #-------------------------------------------------------------------



    @addresses_auto_log_plugin_on(True)
    def on_config(self, config:'MkDocsConfig', **_):
        """
        Extract the config data and compute all the "globally useful" properties for the plugin
        """
#        logger.error(f"\n\n plugin instance id: {id(self)}\n\n")    ##!LOCK
        self.log(LogMessages.on_config_in)
        self._setup(config)
        self._post_setup_config()




    @event_priority(-200)        # pylint: disable-next=unused-argument
    def on_nav(self, _:Navigation=None, *, files:Files, config:MkDocsConfig=None, **__):
        """ Build the references/identifiers for all non markdown files in the docs_dir,
            and log some info for the user to know at what point the executions are.
        """

        if self.plugin_off:
            return

        self._build_identifiers_for_all_files_in_docs_dir_except_markdowns(files)

        self.log(LogMessages.on_nav_in)
#        extractor.refs_snapshot_at('on_nav', self)              ##!





    @addresses_auto_log_plugin_on(
        True, display='page'
    )
    @StaticHandler.on_page_content_cached_with(
        cwd_path_arg_extractor_page_hook,
        never_skip
    )
    def on_page_content(self, html:str, *, page:Page, **_):
        """
        Explore the base html content of the page and extract/build:
            - The references/identifiers defined in the page and their related targets
            - The code snippets (if needed)
            - Store all the identifiers used as targets in the page

        Throw:
            InvalidIdentifierError:     If the identifier is empty ( "{: # oops_space }")
            DuplicateIdentifierError:   If the identifier has already been used (snippets)
            DuplicateIdentifierError:   If the identifier has already been used (references)
        """

        logger.debug('Add file markdown references')
        id_, _ = self.add_links(page, None)                                 ##!
#        extractor.add_html_and_markdown(page, html)                                      ##!
#        id_,_def_link = self.add_links(page, None)                          ##!
#        extractor.add(self, page, id_, '(File on HDD...)', _def_link)       ##!


        logger.debug(f'Add references defined in page: { page.file.src_uri !r}')
        soup = PluginSoupExcludingCodes(html, self)

        for tag in soup.get_tags_with_ids():

            id_   = tag['id']
            is_h  = re.fullmatch(r'h\d+', tag.name)
            kebab = is_h and self.ignore_auto_headers and soup.decompose_header_to_kebab(tag) or None

            logger.debug(f"Handle tag <{tag.name}>, txt={tag.text} (KEBAB={kebab})")
            if kebab == id_:
                self.log_more(f"\033[35mignore_auto_header=true:\033[0m skipped { id_ !r} id.")
                continue

            self.add_links(page, id_)                                       ##!
#            id_,_def_link = self.add_links(page, id_)                       ##>
#            extractor.add(self, page, id_, tag, _def_link)                  ##<


        logger.debug(f'Gather all references used in page: { page.file.src_uri !r}')

        for tag, attr, _, identifier in soup.gen_targeted_addresses_data():
            if VscAutoCompletion.href_is_possible_ref(identifier):
                logger.debug(f'  Reference used: <{ tag.name } { attr }="{ identifier }">')
                self.global_handler.used_refs.add_id(page,identifier)

        all_external_refs = Counter(extract_external_links_refs_in_md(page.markdown))
        refs_in_code_tags = Counter( soup.get_external_refs_in_codes() )
        actual_refs       = all_external_refs - refs_in_code_tags
        for identifier in actual_refs:
            logger.debug(f'  External link reference used: [...][{ identifier }]')
            self.global_handler.used_refs.add_id(page,identifier)






    @event_priority(-200)
    def on_env(self, *_,**__):
        """ * Setup the counter of exceptions (used in fail_fast=False mode)
            * Log the suppressed and added references (either at debug or info level,
              depending on more_verbose)
            * Inform the user before starting the on_page_context actions (because it can be
              long, so better give a hint that something is still happening...)
        """
        if self.plugin_off:
            return

        self._refresh_exception_counter()

        suppressed, added = self.global_handler.references.get_update_refs_definitions_info()

        if suppressed:
            data = ''.join( f'\n  {ref} -> {to}' for ref,to in sorted(suppressed.items()) )
            self.log_more("suppressed references:"+data)

        if added:
            data = ''.join( f'\n  {ref} -> {to}' for ref,to in sorted(added.items()) )
            self.log_more("added references..."+data)


        if self.dump_action != DUMP_ONLY:
            self.log(LogMessages.on_env_in)


#        extractor.refs_snapshot_at('on_env', self)              ##!






    @addresses_auto_log_plugin_on(
        True, display='page', log_errors=False
    )
    @StaticHandler.on_page_context_cached_with(
        cwd_path_arg_extractor_page_hook,
        skip_if_only_dump,
        page_arg_extractor_page_hooks,
    )
    def on_page_context(self, _ctx, *, page:Page, **__):
        """
        Take a peek at the page.content (html), and extract all the links and images, then:
            1. Replace plugin references with the actual relative path (computed accordingly
               to the current "page location to be")
            2. Validate all the addresses (their final/built version, meaning, all the addresses
               in the final html file: either typed manually or built through the plugin), to
               check that a file actually exists. Raise an error otherwise.

        Note: addresses can be "white listed" (one way or another). If they are, their validation
              is skipped and no error will (should...) be thrown.

        Throw during path building:
            LeadingSlashAddressError:   If the target address starts with a slash and isn't
                                        "white listed" (one way or another)
            UnknownIdentifierError:     If an unknown reference identifier is found

        Throw during path checking:
            AddressError:               If use_directory_urls=True, the source file isn't an
                                        index.md file, and the address doesn't start with '../'
            EmptyAnchorError:           If the identifier following an anchor is empty
            ImageWithAnchorError:       If an address used for an image contains an anchor
            AnchorUnknownError:         If an anchor is used but doesn't match with any of the
                                        references stored by the plugin
            AnchorNotInPageError:       If an anchor is used but isn't defined in the current page
            TooManyAnchorsError:        If several anchors are found in the address
            NonRelativeAddressError:    If an address passed the leading slash case but is
                                        not considered relative here (ie starting with a dot)
            NotAnImageAddressError:     If the file is supposed to be an image, but the address
                                        used doesn't match against the imgs_extension_pattern
            FileNotFoundError:          If a file isn't found (image, internal document, or if
                                        the built url, when turned back into a file uri according
                                        to the current configuration doesn't match a markdown file
            TrailingSlashAddressError:  If use_directory_urls is False and an address is ending with
                                        a slash
        """

#        extractor.add_start_ctx(page)             ##!


        to_check = not self.verify_only or page.file.src_uri in self.verify_only

        soup = PluginSoupExcludingCodes(page.content, self)

        for tag, attr, kind, identifier in soup.gen_targeted_addresses_data():
            logger.debug(f'  Build path for: <{ tag.name } { attr }="{ identifier }">')

            if self.fail_fast:
                self._build_address_and_check_it(page, tag, attr, kind, to_check)
            else:
                try:
                    self._build_address_and_check_it(page, tag, attr, kind, to_check)
                except Exception as exc:                                # pylint: disable=broad-exception-caught
                    exc_name = exc.__class__.__name__
                    info     = f"{ plugin_dump_padder(page.file.src_uri) } { tag[attr] }"
                    self.exceptions_counter[exc_name].append(info)

        page.content = str(soup)

        # Search now (see warning below) for external links references that would not have been
        # converted because of a wrong reference: at this point, any `[...][...]` still present
        # in the html content, this is NOT in a code tag, is an invalid external link that has
        # not been replaced by mkdocs.
        #
        # WARNING: this is a destructing operation, on the soup instance!

        soup.destroy_codes()

        markdown_no_code_tags = soup.get_text()
        invalid_external_refs = extract_external_links_refs_in_md(markdown_no_code_tags)

        for wrong_ref in invalid_external_refs:
            if self.fail_fast:
                raise UnknownExternalLinkIdentifierError(
                    f"Unknown identifier { wrong_ref !r}, found in file { page.file.src_uri !r}"
                )
            else:
                exc_name = UnknownExternalLinkIdentifierError.__name__
                info     = f"{ plugin_dump_padder(page.file.src_uri) } { wrong_ref }"
                self.exceptions_counter[exc_name].append(info)

        logger.debug(
            f"on_page_context for page={page.file.src_path!r}: html content update & check done."
        )
#        extractor.add_end_ctx(page)               ##!





    # DO NOT decorate the post build event with cache logic: too much branching logics to
    # apply here, and better to not cache anything (that part won't cost much time anyway!)
    @addresses_auto_log_plugin_on(
        True, log_exceptions_count=True
    )
    def on_post_build(self, *_, **__):
        """ 1) If using VSC, dump all the snippets in the project user snippets file.
            2) Compare the state of the references before the current build and their state
            right now, then warn the user about possible inconsistencies.
            3) raise DumpOnlyError if needed, to abort the build
        """
        if self.dump_action == NO_DUMP:
            self.log(LogMessages.on_post_build_none)
        else:
            self._dump_snippets()

        self._compare_current_state_with_starting_state()

#        extractor.refs_snapshot_at('on_post_build', self)      ##>
#        extractor.flow_dump(self)
#        # extractor.dump(self)                                 ##<

        abort = self.dump_action == DUMP_ONLY
        fatal_if(
            abort, LogMessages.on_post_build_only, DumpOnlyError
        )





    #---------------------------------------------------------------------------




    def log(self, msg:str, with_color=True):
        """ Log with color on the message (unless with_color=False), at INFO level """
        if with_color:
            msg = f"\033[32m{ msg }\033[0m"
        logger.info(msg)



    def log_more(self, msg:str):
        """ Log with color on the message (unless with_color=False), at INFO level """
        show_log = logger.info if self.more_verbose else logger.debug
        show_log(msg)



    def _refresh_exception_counter(self):
        self.exceptions_counter = defaultdict(list)



    #---------------------------------------------------------------------------





    def _setup(self, config:MkDocsConfig):
        """ First inner step of on_config hook (helps with testing) """
        self.config.validate_and_process(config)
        self.global_handler = StaticHandler.inject(self, config)
                            # Inject ONLY after validation/preprocessing of the config
        self._update_watch_list(config)



    @addresses_auto_log_plugin_on(
        step_name="updating the watch list with inclusions and/or external link file",
        log_errors=False,
    )
    def _update_watch_list(self, config:MkDocsConfig):

        has_includes       = self.use_vsc and self.inclusions
        has_external_links = self.external_links_file

        if not has_includes and not has_external_links:
            return

        watching: List[str] = config['watch']
        uri_links = path_manager.to_uri(self.external_links_file) if has_external_links else None

        if has_includes:
            self.log_more(LogMessages.watch_inclusions)

            for inclusion in self.inclusions:
                if not inclusion.is_dir():
                    raise NotADirectoryError(inclusion)
                uri = path_manager.to_uri(inclusion)
                if uri not in watching:
                    watching.append(uri)

        if has_external_links and uri_links not in watching:
            watching.append(uri_links)



    def _post_setup_config(self):
        """ Second inner step of on_config hook (helps with testing) """
        self._extract_and_cleanup_current_snippets_file()
        self._build_inclusions_snippets()
        self._add_external_links_if_needed()






    @addresses_auto_log_plugin_on(
        step_name="extracting the existing code snippets file",
        log_errors=False,
    )
    @StaticHandler.snippets_extraction_cached_with(
        plugin_file_prop_extractor('dump_snippets_file'),
        skip_if_no_vsc_or_no_snippets_file_or_wrong_extension,
    )
    def _extract_and_cleanup_current_snippets_file(self):
        """ If using VSC and a code snippets file is provided, extract it and check if the user
            added anything not related to the plugin. If so, store that for the next dump.
            Every entry related to the plugin is discarded.
        """
        self.log_more(LogMessages.setup_snippets)
        snippets = self.dump_snippets_file.read_text()

        # Keep only things that weren't previously generated (just in case...)
        json_as_dict = json.loads(snippets).items()
        other_snippets = (
            build_other_snippet(name, d)
            for name,d in json_as_dict
            if 'description' not in d or PLUGIN_MARK not in d['description']
        )
        self.global_handler.references.store_other_snippets(other_snippets)




    @addresses_auto_log_plugin_on(
        step_name="building files inclusions snippets",
        log_errors=False,
    )
    def _build_inclusions_snippets(self):
        """ Travel recursively through the content of each paths in the inclusions and build
            the related snippets.
        """
        if self.use_vsc and self.inclusions:

            self.log_more(LogMessages.build_inclusions)

            # seen = set()
            for root_include in self.inclusions:
                for file in root_include.rglob('./*'):
                    if file.is_file():
                    # if file.is_file() and file not in seen:
                        # seen.add(file)
                        self._gather_inclusions_in(file, root_include)


    @StaticHandler.other_files_cached_with(
        cwd_path_arg_extractor_inclusions,
        skip_if_file_already_registered,
    )
    def _gather_inclusions_in(self, file:PathCwd, root_inclusion:Path):
        uri = path_manager.to_uri(file)
        root_inclusion_uri = path_manager.to_uri(root_inclusion)
        self._add_snippets(file, RefKind.Include, uri, uri, root_inclusion=root_inclusion_uri)






    @addresses_auto_log_plugin_on(
        step_name="extracting the external links from a specific file",
        log_errors=False,
    )
    @StaticHandler.external_links_cached_with(
        plugin_file_prop_extractor('external_links_file'),
        skip_if_no_external_links,
    )
    def _add_external_links_if_needed(self):
        """ Add external links

            Note: the references of those links aren't stored, and only the autocompletion
                  informations are actually interesting, because mkdocs will replace the
                  identifiers on it's own. But the equivalent references are stored anyway
                  so that the txt file dumped if use_vsc=false still contains everything.
        """
        self.log_more(LogMessages.setup_links)
        code = self.external_links_file.read_text()
        self._add_external_links(code)


    def _add_external_links(self, code:str):
        """ Unconditional addition of the external links contained in the given markdown code.
            (separated from _add_external_links_if_needed to help with testing)
        """
        matches   = re.findall(r'^\[([^\]]+)\]:\s+(\S+)', code, flags = re.MULTILINE)
        rel_paths = ''.join( f"\n    { id !r}: { link !r}"
                                for id,link in matches
                                if not self.links_white_list_pattern.match(link) )
        fatal_if(
            rel_paths,
            f"All links in the external_links_file ({ self.external_links_file !r}) "
            f"should be absolute, but got:{ rel_paths }\n"
            "(update the configuration property links_white_match, to bypass this).\n",
            NonAbsoluteAddressError
        )
        for identifier, link in matches:
            self.__add_data(self.external_links_file, RefKind.Ext, identifier, link)






    @addresses_auto_log_plugin_on(
        step_name="building the identifiers for all non markdown files in the docs_dir"
    )
    def _build_identifiers_for_all_files_in_docs_dir_except_markdowns(self, files:Files):
        """ Add all fixed files in the docs directory:, except for markdown ones
            (let mkdocs build the urls for those, according to the user's settings)
        """
        self.log(LogMessages.setup_non_md_files)

        for file in files:
            uri   = file.src_uri
            is_md = uri.endswith('.md')
            is_built = (
                uri.startswith('assets/javascripts/')
                or uri.startswith('assets/stylesheets/')
                or uri == 'assets/images/favicon.png'       # mkdocs default config
            )
            cwd_path = self.uni_docs_dir / file.src_path

            if not is_built and not is_md:
                self.add_file_infos(cwd_path, file)




    @StaticHandler.other_files_cached_with(
        cwd_path_arg_extractor_other_files,
        skip_if_file_already_registered,
    )
    def add_file_infos(self, _:PathCwd, file:File):
        """ Add the reference and code snippet for a specific file (non markdown)
            Notes:
                - first argument needed for the other_file_arg_extractor function
                - all files are in the docs_dir, here

            ---

            Only the file existence is needed here, not its content, so "up_to_date" has
            actually no useful meaning for any of this.
            The only reason to go through the decoration/cache thing is to keep track of
            the references of the files, to spot later outdated references in the docs.
        """
        is_img   = self.imgs_extension_pattern.search( file.src_uri )
        kind     = RefKind.Img if is_img else RefKind.File
        cwd_path = self.uni_docs_dir / file.src_uri
        self.__add_data(cwd_path, kind, file.src_uri, file.url)








    def __add_data(
        self,
        src: SourceRef,
        kind: RefKind,
        identifier: str,
        def_link: str,
    ):
        final_id = self._add_snippets(src, kind, identifier, def_link)

        logger.debug(f"Adding reference for {identifier=!r}, using {final_id=!r}, pointing to {def_link=!r}")
        self.global_handler.references.add_id(src, final_id, def_link, identifier)

        return final_id         # investigation purpose only



    def _add_snippets(self, src:SourceRef, kind:RefKind, identifier:str, def_link:str, **extras) -> str :
        """ Add a code snippet to the Reference object.

            @src:        page, or PathCwd to the file generating the snippet
            @kind:       Kind of references (see RefKind)
            @identifier: Bare identifier that will be used to build the code snippet prefix
                         (for example: "plain-address" -> "--plain-address" for a RefKind.Link,
                         or "code_ex.py" -> "<<py code_ex.py" for a RefKind.Include, ...)
            @def_link:   Target document/link for the snippet(s) (as uri)
            @**extras:   Extra arguments that will be passed to the VscAutoCompletion.build_snippet
                         method.
        """
        completer = VscAutoCompletion.get_completer_for(kind)
        final_id  = completer.get_final_identifier(identifier)

        if self.use_vsc:
            logger.debug(f"Adding code snippet(s) for {identifier=!r}: {def_link=!r}")
            for json_id,snippet in completer.build_snippet(identifier, def_link, **extras):
                self.global_handler.references.add_snippet(src, json_id, snippet)

        return final_id




    #-----------------------------------------------------------------------------------





    def add_links(self, page:Page, identifier: Union[str,None]):
        """ Add an id reference to the current References object, building the needed url
            relative to the docs root directory (in the built static site!).
            If no @id is provided or @id is None, build the link for the "bare" file itself.
        """
        logger.debug(f"Adding { identifier !r} reference from page { page.file.src_uri !r}")
        if identifier is None:
            identifier = def_link = page.url or './'
            # Note: page.url can be empty only when using use_directory_url=true

        else:
            # Security check (may happen on mistaken attr: "{: # deep-stuff }") :
            fatal_if(
                not identifier,
                "\n '' (empty) identifier found. Check you didn't make any typo, like "
                "\"{: # oops }\" which has an extra space.",
                InvalidIdentifierError)

            def_link = f"{ page.url }#{ identifier }"

        identifier = self.__add_data(page, RefKind.Link, identifier, def_link)

        return identifier, def_link         # investigation purpose only



    def __build_path_from_caller_to_id(self, caller:Page, identifier:str):
        """ Given an identifier and the Page from where it's used, build the appropriate
            relative path for the built site and returns it.
        """
        refs = self.global_handler.references
        fatal_if(
            not refs.has_id(identifier),
            f"Unknown identifier { identifier !r}, found in file { caller.file.src_uri !r}",
            UnknownIdentifierError
        )

        n_upper       = caller.url.count('/')
        up_to_docs    = '../' * n_upper or './'
        relative_link = up_to_docs + refs.get_ref_target_address(identifier)

        logger.debug(f"Built relative path for {identifier=!r}: {relative_link=!r}")
        return relative_link




    @addresses_auto_log_plugin_on(
        display="page:0 tag:1 attr:2", joiner=' ',
        step_name="Replacing the identifier with the relative address and checking their validity",
    )
    def _build_address_and_check_it(
        self,
        page: Page,
        tag: Tag,
        attr: str,
        kind: AddressKind,
        to_check: bool,
    ):
        identifier = target = tag[attr]
        is_built_ref = False

        # Forbid accidental addresses with leading slash:
        fatal_if(
            to_check and identifier.startswith('/'),
            f"No leading slash address allowed: {identifier!r}.\nIf you absolutely need it,"
            ' "white list it", one way or another (note: if using the white list matcher,'
            ' you should use a very strict/long match rule, so that this verification '
            'cannot become obsolete/useless).',
            LeadingSlashAddressError
        )

        if VscAutoCompletion.href_is_possible_ref(identifier):
            logger.debug(f"Build the correct relative path for \"{identifier}\"...")
            target    = self.__build_path_from_caller_to_id(page, identifier)
            tag[attr] = target
            logger.debug(f"\"{ identifier }\" built as: { target !r}")
            is_built_ref = True


        # If not bare anchor and not any obvious "relative" indications, enforce the
        # relative aspect of the target, because that's how mkdocs will finally handle
        # the address in the built site (note: this ahs been added in the logic because
        # the leading dot might be lost on images addresses, specifically...):
        if not target.startswith('.') and not re.match('[/#]', target):
            target = './' + target
            logger.debug(f"Enforced relative path to: {target=!r}")

        # Check the address only if it has not been built by the plugin. This is assuming
        # it will do a proper job... :p )
        if to_check and not is_built_ref:
            AddressChecker.check(self, page, target, kind)




    def _dump_snippets(self):
        logger.debug(f"Building data to dump (use_vsc={ self.use_vsc })")

        if self.use_vsc:
            code = self.global_handler.references.build_json_snippets()
        else:
            items_gen = self.global_handler.references.gen_refs_items()
            code = "".join(
                f"{ plugin_dump_padder(identifier+':') } { file } \n"
                for identifier,file in sorted(items_gen, key=itemgetter(1))
            )

        self.dump_snippets_file.parent.mkdir(exist_ok=True)
        self.dump_snippets_file.write_text(code)
        self.global_handler.files_tracker.mark_file(self.dump_snippets_file)

        self.log(LogMessages.on_post_build_out)



    @addresses_auto_log_plugin_on(
        step_name="comparing references defined on last run with the current one",
        log_errors=False,
    )
    def _compare_current_state_with_starting_state(self):
        self.log(LogMessages.cross_ref_check)
        oops_lst = self.global_handler.get_missing_refs_infos()
        fatal_if(
            oops_lst,
            "References found in existing documents that do not point anywhere (the reference "
            +"isn't defined anymore):"
            +''.join(oops_lst),
            OutdatedReferenceError
        )



"""
Centralize all the elements needed to improve the overall performances, using a global
cache to:
    - Register all the references definitions ("!!img_png", "--header-link", ...) and
      where they are defined (file).
    - Register all the usages of each references (in what file they are used, and what
      file is the target).
    - Keep track of what files got deleted/modified since last build (relative to one
      "mkdocs serve" command).
With all this, the plugin will check the minimal number of files necessary on a build.
"""

from .addresses_usage_pool import AddressesUsagePool
from .references import References
from .static_handler import StaticHandler


import re
import json
import math
from typing import Dict

from .exceptions import AddresserError







def fatal_if(truthy, msg:str, exc=AddresserError):
    if truthy:
        raise exc(msg)




def plugin_dump_padder(k:str, pad=25):
    return k.ljust(pad * math.ceil( len(k) / pad ))


def build_other_snippet(name:str, dct:Dict):
    return f'"{name}": { json.dumps(dct, indent=4) }'


def extract_external_links_refs_in_md(markdown:str):
    """ Extract all the identifiers of externals links in the given markdown code:
                [...][identifier]
    """
    return re.findall(r'\[[^\]]*\]\[([^\]]+)\]', markdown)



# pylint: disable=unused-private-member

import os

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any
from enum import IntEnum, auto

from mkdocs.structure.pages import Page

from mkdocs_addresses import path_manager
from mkdocs_addresses.vsc_auto_completion_handler import RefKind, VscAutoCompletion


from .exceptions import (
    AbstractAddressError,
    AnchorNotInPageError,
    AnchorUnknownError,
    EmptyAnchorError,
    ImageWithAnchorError,
    MarkdownLinkError,
    NoMoveUpDirOnNonIndexPageUrlsAsDirAddressError,
    NonRelativeAddressError,
    NotAnImageAddressError,
    TooManyAnchorsError,
    TrailingSlashAddressError,
)

if TYPE_CHECKING:
    from .addresses_plugin import AddressAddressesPlugin





class AddressKind(IntEnum):
    """ Enum representing the different kind of attribute/addresses that will be
        handled through the plugin.
    """
    HREF = auto()       # <a href=...> => link
    SRC  = auto()       # <img src=.../> => img




@dataclass
class AddressChecker:
    """
    Class centralizing the logic to check the validity of various urls in the final html pages
    """

    plugin:     'AddressAddressesPlugin' # Plugin instance...
    page:        Page                    # Page from where the link address is used
    address:     str                     # The address to validate, as shown in the html code
    kind:        AddressKind             # Identify the context in which the address is used
    #-----------------------------
    # auto-computed in post init:
    has_trailing_slash: bool = None      # True if the address ends with '/'
    is_html_file:bool = None             # True if the address ends with '.html(#...)?'
    caller_dir:  str  = None             # Current (relative) directory containing the source file
    src_address: str  = None             # Debugging purpose
    univ_address:Path = None             # OS dependent version of the given address to validate
    full_address:Path = None             # OS dependent version of the real path of the address
    anchor:      str  = None             # Any anchor spotted in the address


    def __post_init__(self):
        self.src_address  = self.address        # archive for debugging purpose

        # Prevent little jokes... (and may become useful for "wrongish use" of mkdocstrings...)
        if self.address.startswith('./#'):
            self.address = self.address[2:]


        self.__validate_hashtag_and_reassign_address_and_anchor(self.address)
        if self.address is None:
            # Actually found a bare anchor (yeah, not obvious, I know...)
            return

        self.has_trailing_slash = self.address.endswith('/')
        self.is_html_file       = self.address.endswith('.html')
        self.caller_dir         = os.path.dirname(self.page.file.src_path)


        if self.plugin.use_directory_urls and not self.page.file.src_uri.endswith('index.md'):
            # NOTE: bare anchors have already been handled (see above) and are not subject to
            #       use this branch of logic.
            # The paths are "computed" according to the final built site, but when checking
            # validity of the addresses, the concrete dev version must be used.
            # This introduces a discrepancy when use_directory_urls is true and the current
            # file is not an index.md file: the current file will be converted to a directory
            # containing its own index.html file. This extra layer is "included" in the logic
            # building the relative paths, which means all relative paths from a "non index.md"
            # file will generate an extra `../` that must be removed, before checking the
            # address.
            # This is done by removing one leading dot, which allow to keep the consistency for
            # the checks about "a valid link should be relative, hence start with a dot" done
            # later in the code.

            self.__fatal_if(
                not self.address.startswith('../'),         # security check (used in the tests)
                "Address from a non index.md file should always have in a leading '../' but got: "
                f"{ self.src_address !r}",
                NoMoveUpDirOnNonIndexPageUrlsAsDirAddressError
            )
            # Remove one leading fot to match the actual tree hierarchy at build time:
            self.address = self.address[1:]

        self.univ_address = path_manager.to_os(self.address)
        full_address      = self.plugin.uni_docs_dir / self.caller_dir / self.univ_address
        self.full_address = path_manager.to_os(os.path.realpath(full_address))
            # WARNING: the above shenanigan (Path -> os -> Path) is needed because on Windows,
            # Path.resolve may fail if an url reaches this code (if it's not white listed)

    def __validate_hashtag_and_reassign_address_and_anchor(self, address:str):

        # Addresses for img.src stay unchanged (just check no anchors are in there...)
        if self.kind == AddressKind.SRC:
            self.__fatal_if(
                '#' in address, "img.src address should never contain '#'",
                ImageWithAnchorError,
            )
            self.address = address

        # Handle direct links to anchors in the current file itself:
        elif address.startswith('#'):
            self.__validate_anchor(address[1:])
            self.address = None         # will indicate that the address is actually valid

        else:
            # Handle possible trailing anchor:
            lst = address.split('#')
            if len(lst)==2:
                self.address, self.anchor = lst
                self.__fatal_if(
                    not self.anchor,
                    f"Invalid link: Empty anchor identifier: { address !r}",
                    EmptyAnchorError,
                )
                self.__validate_anchor(self.anchor, True)

            self.__fatal_if(
                len(lst)>2,
                f"Invalid link: contains more than one anchor: { address !r}",
                TooManyAnchorsError,
            )


    def __str__(self):
        return f"""AddressChecker data:
    { self.page.file.src_uri =}
    { self.page.url =}
    { self.caller_dir =}
    { self.src_address =}
    { self.has_trailing_slash =}
    { self.is_html_file =}
    { self.address =}
    { self.univ_address =}
    { self.full_address =}
    cwd = { os.getcwd() }
"""


    @classmethod
    def check(cls, plugin:'AddressAddressesPlugin', page:Page, address:str, kind:AddressKind):
        """
        Entry point, providing all the needed data, then defining the validations to do and
        performing them.

        WARNING:
            It's important to keep in mind that, at this point, the addresses in the html
            are all written in a relative way, hance might be like "../../assets/...)".
            It's only in the built site that they are shown as "resolved"/absolute.

        @throws: AddresserError, subclasses of AddresserError, FileNotFoundError.
        """

        checker = cls(plugin, page, address, kind)     # see __post_init__ validations...

        if checker.address is None:
            return     # valid bare hashtag identifier, kind cannot be IMG, here.

        is_not_relative = not checker.address.startswith('.')
        checker.__fatal_if(
            is_not_relative,
            f"{ checker.address !r} should be relative path (starting with at least one dot)",
            NonRelativeAddressError,
        )

        is_md = checker.full_address.suffix == '.md'
        checker.__fatal_if(
            is_md,
            "Links to markdown files should never be used in the final documentation. "
            "If this is intended, white list this address in the plugins configuration.",
            MarkdownLinkError
        )

        # Pick the validation logic to use:
        if checker.kind == AddressKind.SRC:
            checker.__validate_img_src_attr()

        elif checker.plugin.use_directory_urls:
            checker.__validate_with_docs_as_dirs()

        else:
            checker.__validate_with_docs_as_html()



    def __validate_anchor(self, anchor:str, allow_external_to_file=False):
        completer  = VscAutoCompletion.get_completer_for(RefKind.Link)
        identifier = completer.get_final_identifier(anchor)
        references = self.plugin.global_handler.references
        unknown_id = not references.has_id(identifier)
        self.__fatal_if(
            unknown_id,# and not allow_external_to_file,
            f'Unknown anchor: "#{ anchor }"',
            AnchorUnknownError
        )
        not_in_page = not references.is_identifier_in_source(identifier, self.page)
        self.__fatal_if(
            not_in_page and not allow_external_to_file,
            f'Anchor: "#{ anchor }" is not defined in the current page',
            AnchorNotInPageError
        )



    def __validate_img_src_attr(self):
        is_file = self.full_address.is_file()     # check first (testing purpose)
        self.__fatal_if(
            not is_file,
            f"{ self.full_address !r}",
            FileNotFoundError
        )
        is_img  = self.plugin.imgs_extension_pattern.search(self.full_address.name)
        self.__fatal_if(
            not is_img,
            f"{ self.full_address !r} is not identified as an image",
            NotAnImageAddressError
        )



    def __validate_with_docs_as_dirs(self):
        # (note: anchors (if any) have already been validated)

        is_file         = self.full_address.is_file()
        actually_dir    = self.full_address.is_dir()

        maybe_doc_file  = self.full_address.parent / (self.full_address.stem + '.md')
        actually_doc_md = maybe_doc_file.is_file()

        maybe_index     = self.full_address / 'index.md'
        actually_index  = maybe_index.is_file()

        if not self.has_trailing_slash:
            # HAS to be an existing file, then:
            self.__fatal_if(not is_file, self.full_address, exc=FileNotFoundError)

        elif actually_dir:
            # If it's a directory, this one has to contain a index.md file
            self.__fatal_if(
                not actually_index,
                f"{ maybe_index } should not point to an actual directory. If this is intended, "
                "white list this address",
                exc=FileNotFoundError
            )

        else:
            self.__fatal_if(not actually_doc_md, maybe_doc_file, exc=FileNotFoundError)



    def __validate_with_docs_as_html(self):
        # (note: anchors (if any) have already been validated)

        is_dir  = self.full_address.is_dir()
        is_file = self.full_address.is_file()

        # Forbid pointing to an actual directory in "html mode":
        self.__fatal_if(
            is_dir or self.has_trailing_slash,
            f"{self.full_address!r} address is not supposed to be a directory, if mkdocs "
            "use_directories_url is false. If it's intended, add this address to the white list.",
            TrailingSlashAddressError if self.has_trailing_slash else FileNotFoundError
        )

        if not is_file:
            # If the address points to a "currently non existent file" on the disk, while the
            # address is an html one, the address is always wrong:
            self.__fatal_if(not self.is_html_file, self.full_address, FileNotFoundError)

            # If the uri points to an html file and that file has not been actually found, it HAS
            # to be a markdown documentation file:
            maybe_doc_file = self.full_address.parent / (self.full_address.stem + '.md')
            is_md          = maybe_doc_file.is_file()
            self.__fatal_if(not is_md, maybe_doc_file, FileNotFoundError)





    def __fatal_if(self, truthy, msg:Any, exc=AbstractAddressError):
        if truthy:
            tail = f'\n(Source address: {self.src_address!r} in file {self.page.file.src_uri!r})'
            raise exc(str(msg) + tail)




import json

from typing import TYPE_CHECKING, Dict, Iterable, Set, Union
from dataclasses import dataclass, field
from collections import defaultdict

from mkdocs_addresses.static_handler.page_data_tracker import PageDataTracker

from .types_aliases import SourceRef, PathCwd


if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin




class AddressesUsagePool(PageDataTracker):
    """ Keep track of the sources files (as uris) that are using references, like:

            ![alt](!!logo_svg)

        References can only be _used_ in pages, so the sources are the page.file.src_uri.
        No need for OSCwdPath thingy or so... BUT, to get a homogeneous behavior of the
        whole plugin, store the cwd rooted paths instead.
    """

    _old_id_to_sources: Dict[str,Set[PathCwd]]
    _old_source_to_ids: Dict[PathCwd,Set[str]]


    def __init__(self, uni_docs_dir:PathCwd):
        super().__init__(uni_docs_dir)
        self._old_id_to_sources = {}
        self._old_source_to_ids = {}


    #---------------------------------


    def archive_current(self, uni_docs_dir:PathCwd):
        """ Dump a deepcopy of the to internal structures: (id_to_sources, source_to_ids) """
        super().archive_current(uni_docs_dir)
        self._old_id_to_sources = self.id_to_sources.copy()
        self._old_source_to_ids = self.source_to_ids.copy()


    def get_missing_refs_infos(self, currently_defined_ids: Iterable[str]):
        """ Compare the archived data references the current ones, to find files that use outdated
            references, and returns a list of strings describing those.
            Needed because the references in pages that are "up to date" won't be checked during
            the on_page_context hook.
        """
        missings = set(self._old_id_to_sources) - set(currently_defined_ids)
        err_msgs = [
            f"\n    { ref !r} { MIDDLE_MSG } { src }"
                for ref in sorted(missings)
                for src in self._old_id_to_sources[ref]
        ]
        return err_msgs


MIDDLE_MSG = "used in file"  # this is used in the tests to cut properly a line into the original information



from mkdocs.exceptions import PluginError


class AddresserError(PluginError):
    """ Generic mkdocs-addresses exception. Extends PluginError """



#------------------------------------------------



class AbortError(AddresserError):
    """ Thrown at the face of mkdocs logger... :p
        (see addresses_auto_log final try/except block)
    """


class AddresserConfigError(AddresserError):
    """ Plugin config related error """



class InvalidOperationError(AddresserConfigError):
    """ Don't to that... """



class DumpOnlyError(AddresserError):
    """ Generic mkdocs-addresses exception. """



class MarkdownLinkError(AddresserError):
    """ Generic mkdocs-addresses exception. """



#------------------------------------------------



class AbstractAddressError(AddresserError):
    """ Path/address related errors. Extends AddresserError """


class NonAbsoluteAddressError(AbstractAddressError):
    """ Address is not considered absolute. """

class NonRelativeAddressError(AbstractAddressError):
    """ Address is not considered relative (no leading dot) """

class LeadingSlashAddressError(AbstractAddressError):
    """ Addresses with leading slash are not allowed """

class TrailingSlashAddressError(AbstractAddressError):
    """ Addresses with trailing slash """

class NotAnImageAddressError(AbstractAddressError):
    """ The address isn't considered the one of an image. """

class NoMoveUpDirOnNonIndexPageUrlsAsDirAddressError(AbstractAddressError):
    """ In a non index.md file while using use_directory_urls, the address does not start
        with a "../"
    """




class AnchorAddressError(AbstractAddressError):
    """ Invalid anchor found in an Address """


class ImageWithAnchorError(AnchorAddressError):
    """ Images src attribute should never contain an anchor """

class TooManyAnchorsError(AnchorAddressError):
    """ More than one hashtag in the address """

class AnchorUnknownError(AnchorAddressError):
    """ Anchor not referenced anywhere """

class AnchorNotInPageError(AnchorAddressError):
    """ Anchor not referenced in the current page """

class EmptyAnchorError(AnchorAddressError):
    """ Trailing anchor without identifier (spot {: # wrong } errors) """



#------------------------------------------------



class IdentifierError(AddresserError):
    """ References/Id related errors """


class InvalidIdentifierError(IdentifierError):
    """ A reference identifier is invalid """

class DuplicateIdentifierError(IdentifierError):
    """ An identifier has already been registered """

class UnknownIdentifierError(IdentifierError):
    """ Unknown identifier """

class UnknownExternalLinkIdentifierError(IdentifierError):
    """ Unknown identifier """



class OutdatedReferenceError(IdentifierError):
    """ Identifier used in some document, but that isn't defined anywhere anymore (note: can occur
        only for unchanged documents that were cached and so, not checked on this build, while the
        reference they are using has been deleted in another file).
    """



#-------------------------------------------------

class ConcurrentPluginError(Exception):
    """ Using only for testing purpose, ensuring that pytest doesn't run in // """




import re
from typing import Generator, Set, NamedTuple, TYPE_CHECKING

from bs4 import BeautifulSoup
from bs4.element import Tag

from mkdocs_addresses.toolbox import extract_external_links_refs_in_md


from .addresses_checker import AddressKind
from .vsc_auto_completion_handler import RefKind


if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin



class SoupTag(NamedTuple):
    tag: Tag
    attr: str
    kind: RefKind
    identifier: str



TR_FR_TABLE = str.maketrans(
    "àâäéèêëîïôöùûüç",
    "aaaeeeeiioouuuc"
)

def kebabize_fr(msg:str):
    ke     = msg.lower().translate(TR_FR_TABLE)
    purged = re.sub(r'&(amp|lt|gt);|[^a-z0-9_ -]+', '', ke)
    bab    = re.findall(r'\w+', purged)
    return '-'.join(bab)



class PluginSoupExcludingCodes(BeautifulSoup):
    """ Personalized version of the BeautifulSoup parser, it will ignore the tags that are
        contained in <code> tags.

        Takes the plugin as argument in the constructor so that it can extract the configuration
        data from it.
    """

    __codes: Set[Tag]
    __plugin: 'AddressAddressesPlugin'


    def __init__(self, html:str, plugin:'AddressAddressesPlugin'):
        super().__init__(html, features="html.parser")
        self.__codes = [*super().find_all('code')]
        self.__plugin = plugin



    def decompose_header_to_kebab(self, header:Tag):
        """ Takes an header tag (<h...>) and converts it to its string content, by removing
            all the unwanted children.

            WARNING: THIS IS MUTATING THE TREE!
        """
        for child in header:
            if isinstance(child,Tag) and "headerlink" in child.get('class',''):
                child.decompose()

        stripped   = header.get_text()
        keb_header = kebabize_fr(stripped)
        return keb_header



    def find_all(self,*a,**kw) -> Set[Tag] :
        candidates = super().find_all(*a,**kw)
        exclusions = { tag for code in self.__codes
                           for tag in code.find_all(*a,**kw) }

        # Filter output as list, and not using sets, to keep the iteration order consistent
        # in other parts of the program:
        return [ cnd for cnd in candidates if cnd not in exclusions ]



    def get_tags_with_ids(self):
        """ Extract all the tags in the html content that:
                - Contain an id attribute (potentially defined on the fly by mkdocs: headers!)
                - Are not in a <code> tag
                - Compliant with the current plugin configuration (ignored ids and classes,
                  white lists, ...)
        """
        return self.find_all(
            id=self.html_id_to_gather,
            class_=self.html_class_to_gather,
        )


    def gen_targeted_addresses_data(self) -> Generator[SoupTag,None,None]:
        """ Explore an html content, searching for links href and images src addresses,
            and yield the corresponding tag with some other informations.
            Note: anchors definitions are skipped.
        """
        links = self.find_all('a', class_=self.html_class_to_gather)
        imgs  = self.find_all('img')

        targeted_tags = (
            (links, 'href', AddressKind.HREF),
            (imgs,  'src',  AddressKind.SRC),
        )
        for tags,attr,kind in targeted_tags:
            for tag in tags:
                if not tag.has_attr(attr):
                    continue
                identifier = tag[attr]
                if not self.is_identifier_or_link_to_ignore(identifier):
                    yield SoupTag(tag, attr, kind, identifier)




    def is_identifier_or_link_to_ignore(self, id_or_address:str):
        """ Check if the given identifier or address must be ignored or not """
        return (
           id_or_address in self.__plugin.links_white_list
           or id_or_address in self.__plugin.ignored_ids
           or self.__plugin.links_white_list_pattern.match(id_or_address)
        )


    def html_class_to_gather(self, class_:str):
        if not class_:
            return True
        kls   = set(class_.split())
        inter = self.__plugin.ignored_classes & kls
        return not inter



    def html_id_to_gather(self, html_id:str):
        return (
            html_id is not None
            and not html_id.startswith('__')
            and not self.is_identifier_or_link_to_ignore(html_id)
        )



    def destroy_codes(self):
        """ Mutate the current instance, removing all the <code> tags """
        for code in self.__codes:
            code.decompose()

    def get_external_refs_in_codes(self):
        """ Generate all the external references that are in code tags (hence, have not been
            modified in the html)
        """
        return (
            id_ for code in self.__codes
                for id_ in extract_external_links_refs_in_md(code.get_text())
        )



import json
from pathlib import Path

from typing import TYPE_CHECKING, Dict, Iterable, Set
from collections import defaultdict
from abc import ABCMeta

from mkdocs_addresses.exceptions import DuplicateIdentifierError

from .types_aliases import SourceRef, PathCwd


if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin



class PageDataTracker(metaclass=ABCMeta):
    """ Abstract class providing the general logic for data structures that are tracking
        information contained in some docs pages.

        Only holds the general behaviors, assuming the subclass will always have  at least
        those two data structures:

            id_to_sources: Dict[str,Set[str]]
            source_to_ids: Dict[PathCwd,Set[str]]
    """
    UNIQUE_SRC: bool = False        # can be registered once only

    _uni_docs_dir: PathCwd

    id_to_sources: Dict[str,Set[str]]
    source_to_ids: Dict[PathCwd,Set[str]]


    def __init__(self, uni_docs_dir:Path):
        self._uni_docs_dir = uni_docs_dir
        self.id_to_sources = defaultdict(set)
        self.source_to_ids = defaultdict(set)


    def __str__(self):
        return json.dumps({
            "id_to_sources": {id:list(vs) for id,vs in self.id_to_sources},
            "source_to_ids": {id:list(vs) for id,vs in self.source_to_ids}
        }, indent=2)


    #------------------------------------------------------


    def format_source(self, source:SourceRef) -> PathCwd:
        """ Converts the given source information into a Path rooted at cwd.
            If the input is not a mkdocs Page, assume it's already given appropriately rooted.
        """
        # Does NOT check against Page class because of the fake Pages used in the tests.
        src = (
               source if isinstance(source,Path) else
               Path(source) if isinstance(source,str) else
               self._uni_docs_dir / source.file.src_uri
        )
        return src


    #------------------------------------------------------


    def add_id(
        self,
        source:         SourceRef,
        identifier:     str,
        id_declared_as: str = None
):
        """ Store the given data, associated to the given id, for the given page.
            Raise DuplicateIdentifierError if @id has already been used before while
            self.UNIQUE is True.
            Returns the source (as PathCwd) so that a child class can store more additional
            logic if needed.
        """
        cwd_path   = self.format_source(source)
        known_srcs = self.id_to_sources[identifier]

        if self.UNIQUE_SRC and known_srcs:
            there = next(iter(known_srcs))
            info  = "" if id_declared_as is None else f"(found written as { id_declared_as !r})"
            raise DuplicateIdentifierError(
                f"{ identifier !r} is already used.\n"
                f"  - Attempt to add it to references\n"        # this line is used in the tests: do not modify
                f"  - Adding from file { cwd_path }{ info }\n"
                f"  - Already registered with file {there}\n"
            )

        known_srcs.add(cwd_path)
        self.source_to_ids[cwd_path].add(identifier)

        return cwd_path



    def remove_source(self, source:SourceRef):
        """ Remove all the informations related to the given page from the data structure.
            Does nothing if the page isn't registered.
            Returns the PathCwd involved, as well as the corresponding ids to the caller.
        """
        src = self.format_source(source)
        ids: Iterable[str] = self.source_to_ids.pop(src, ())
        for identifier in ids:
            sources_pool = self.id_to_sources[identifier]
            sources_pool.discard(src)
            if not sources_pool:
                del self.id_to_sources[identifier]

        return src, ids



    def remove_id(self, identifier:str):
        """ Remove all the informations related to the given identifier from the data structure.
            Does nothing if the reference isn't registered.
            Returns the sources where the identifier was found.
        """
        sources: Iterable[str] = self.id_to_sources.pop(identifier, ())
        for source in sources:
            ids_pool = self.source_to_ids[source]
            ids_pool.discard(identifier)
            if not ids_pool:
                del self.source_to_ids[source]
        return sources


    #------------------------------------------------------


    def has_id(self, identifier:str):
        return identifier in self.id_to_sources

    def has_source(self, source:SourceRef):
        src = self.format_source(source)
        return src in self.source_to_ids

    def is_identifier_in_source(self, identifier:str, source:SourceRef):
        """ Tells if the given identifier has been/is still declared in the given source """
        cwd_src = self.format_source(source)
        return identifier in self.source_to_ids[cwd_src]





    def get_sources_with_id(self, identifier:str):
        """ Extract all filenames using the given @identifier """
        return self.id_to_sources[identifier]

    def get_ids_in_source(self, source:SourceRef):
        """ Extract all identifiers used the given @page (Page or filename) """
        src = self.format_source(source)
        return self.source_to_ids[src]


    #---------------------------------


    def archive_current(self, uni_docs_dir:PathCwd):
        """ Only takes care of the uni_docs_dir logistic part. Actual archiving must be done
            by the child class .
        """
        self._uni_docs_dir = uni_docs_dir




import os

from dataclasses import dataclass, field
from typing import Dict

from mkdocs_addresses import path_manager
from mkdocs_addresses.logger import logger

from .types_aliases import PathCwd, UriCwdPathStr







@dataclass
class FilesTracker:
    """ Somewhat "augmented" version of mkdocs.Files, tracking the source files and their date
        of last modification:
                - path (OS, relative to cwd) -> time
    """

    real_path_to_time: Dict[PathCwd,float] = field(default_factory=dict)



    def mark_file(self, source:PathCwd):
        self.real_path_to_time[source] = source.stat().st_mtime


    def is_file_up_to_date(self, source:PathCwd):
        """ Check if the source file has been updated since the last build.
            NOTES:
                - the file HAS to exist, here.
                - the path must be "os", and given relatively to the cwd
        """
        is_cached     = source in self.real_path_to_time
        cached_at     = is_cached and self.real_path_to_time[source]
        still_here    = is_cached and source.is_file()
        current       = still_here and source.stat().st_mtime
        is_up_to_date = is_cached and cached_at == current

        if not is_cached:
            logger.debug(f"Not cached yet: {source}")
        elif not is_up_to_date:
            logger.debug(f"File to update: {source}")

        return is_up_to_date


    def remove_as_cwd_uri(self, uri:UriCwdPathStr):
        """ Untrack a file, given its uri path (relative to cwd) """
        real = path_manager.to_os(uri)
        del self.real_path_to_time[real]


    def remove_as_cwd_path(self, real:PathCwd):
        """ Untrack a file, given its OS path (relative to cwd) """
        del self.real_path_to_time[real]


    def get_uris_and_paths(self):
        """ Returns a tuple fo pairs (CachePairPath) for all the tracked files,
            relative to the cwd.
        """
        return tuple(self.real_path_to_time)



"""
See: https://github.com/mkdocs/mkdocs/discussions/3241
"""
from argparse import Namespace
import traceback
import logging

from typing import TYPE_CHECKING, Any, MutableMapping, Tuple
from functools import wraps

from mkdocs_addresses.exceptions import AbortError, AddresserError
from mkdocs_addresses.toolbox import plugin_dump_padder

if TYPE_CHECKING:
    from .addresses_plugin import AddressAddressesPlugin






class LogMessages(Namespace):
    """ Holds all the INFO level messages """

    on_config_in        = "Setup config and plugin instance"
    setup_non_md_files  = "Add references for non md files in docs_dir"
    on_nav_in           = "Gather references definitions in all pages..."
    on_env_in           = "Replace references with addresses, checking them..."
    on_post_build_none  = "No code snippets created"
    on_post_build_out   = "Code snippets dump successful"
    on_post_build_only  = '"Code snippets only" mode: abort mkdocs build.'
    cross_ref_check     = "Check for outdated references in unmodified files"

    # --- log_more ---
    setup_snippets      = "Extract existing code snippets"
    watch_inclusions    = "Add inclusions directories and/or external links file to mkdocs watch array"
    build_inclusions    = "Build code snippets for inclusions"
    setup_links         = "Extract external links"









class Logger(logging.LoggerAdapter):
    """A logger adapter to prefix messages with the originating package name."""

    is_in_tests = False

    def __init__(self, prefix: str, logger: logging.Logger):
        """Initialize the object.

        Arguments:
            prefix: The string to insert in front of every message.
            logger: The logger instance.
        """
        super().__init__(logger, {})
        self.prefix = prefix

    def process(self, msg: str, kwargs: MutableMapping[str, Any]) -> Tuple[str, Any]:
        """Process the message.

        Arguments:
            msg: The message:
            kwargs: Remaining arguments.

        Returns:
            The processed message.
        """
        return f"{self.prefix}: {msg}", kwargs








def get_plugin_logger(name: str) -> Logger:
    """Return a logger for plugins.

    Arguments:
        name: The name to use with `logging.getLogger`.

    Returns:
        A logger configured to work well in MkDocs,
            prefixing each message with the plugin package name.
    """
    _logger = logging.getLogger(f"mkdocs.plugins.{name}")
    return Logger(name.split(".", 1)[0], _logger)




logger = get_plugin_logger(__name__.replace('_','-'))



def addresses_auto_log_plugin_on(
    is_event=False,
    display:str="",
    joiner='\n  ',
    step_name:str=None,
    log_errors:bool=True,
    log_exceptions_count:bool=False,
):
    """ Wrap a method of AddressAddressesPlugin and automatically log:
            - DEBUG: When entering the decorated method
            - ERROR: the raised error with custom informations:
                - the error kind
                - the method name
                - info about the arguments (see @display)

        @is_event:  True if the decorated method is one of the plugin event method (identify
                    entry points, to not raise from there, if fail_fast option is false).
        @display:   A space separated string in the format "name" or "name:index", which tells
                    what to add to the log message, and how to find it in the varargs/kwargs:
                       "name" => search in kwargs / "name:i" => use varargs[int(i)]

                    WARNING about arguments with default values: they are in the signatures, but
                    aren't showing up in the *a and **kw of the decorators! => simplest way out of
                    this: just pass the default value explicitly (or avoir default values...).
        @joiner:    String to use to join the "extras" infos (see @display)
        @step_name: if defined, replace the method name in the log, when an error is logged.
        @log_errors: allow to not log when there is already a nested call logging stuff, or
                    when an upper level will do it

        The arguments are transformed using the functions in TRANSFORM, associating a
        transformation to a name. By default, not transformation are done.
    """

    display = display and display.split()
    transforms = {
        'page': lambda page: repr(page.file.src_uri),
        'tag':  lambda tag:  repr(tag.name),
    }

    def build_info_getter(prop:str, i:str=None):
        trans = transforms.get(prop, repr)
        if i is None:
            return lambda _,kw: f"{ prop }={ trans(kw[prop]) }"
        else:
            return lambda a,__: f"{ prop }={ trans(a[int(i)]) }"

    extras = [ build_info_getter(*chunk.split(':')) for chunk in display ]

    def decorator(method):
        @wraps(method)
        def wrapper(self:'AddressAddressesPlugin', *a, **kw):

            if self.plugin_off:
                return

            # pages context exceptions global count must be shown before the information about
            # entering on_post_build is given to the user:
            if log_exceptions_count and self.exceptions_counter:

                sorted_items = sorted(self.exceptions_counter.items(), key=lambda it: -len(it[1]))
                warn_message = (
                    'running with fail_fast=false...\n' +
                    f'Summary of the exceptions raised handling the addresses ni the pages:' +
                    ''.join(
                        f'\n\033[31m    in {plugin_dump_padder(exc_name+":")} { len(files) } \033[0m'
                        + "".join( f"\n      {file}" for file in files)
                        for exc_name,files in sorted_items
                    )
                )
                logger.warning(warn_message)


            more_info  = joiner.join( get_info(a,kw) for get_info in extras)
            tail_entry = more_info and " with " + more_info
            where      = step_name or method.__name__

            more_verbose   = getattr(self, 'more_verbose', False)     # fails for non obvious reason => default to False...
            entry_log_func = logger.info if is_event and more_verbose else logger.debug
            entry_log_func(f"\033[34m{ where }{ tail_entry }\033[0m")

            try:
                return method(self, *a, **kw)

            except Exception as exc:                        # pylint: disable=all
                if log_errors:
                    exc_name      = exc.__class__.__name__
                    intro         = ( f"\033[31m{ exc_name }\033[0m - { where }(...)"
                                    +  "\n  " * bool(more_info) )
                    formatted_exc = f"\033[34m{ exc }\033[0m"
                    log_msg       = f"{ intro }{ more_info }\n  { formatted_exc }"
                    if not isinstance(exc, AddresserError):
                        log_msg += '\n' + traceback.format_exc()

                    logger.error(log_msg)

                if self.fail_fast or not is_event:
                    if not logger.is_in_tests and self.fail_fast and not log_errors and is_event:
                        # here, the error already got logged to the console, but mkdocs will log
                        # it's message again with its own logger. So throw something else to avoid
                        # the duplicated message
                        raise AbortError('') from exc
                    else:
                        raise

        return wrapper
    return decorator



#from threading import Lock      ##!LOCK


from typing import TYPE_CHECKING, Any, Callable

from functools import wraps

from mkdocs.structure.pages import Page
from mkdocs.config.defaults import MkDocsConfig


from mkdocs_addresses.logger import logger
from mkdocs_addresses.exceptions import (
    AddresserError,
#    ConcurrentPluginError,                                 ##!LOCK
)


from .files_tracker import FilesTracker
from .types_aliases import PathCwd
from .addresses_usage_pool import AddressesUsagePool
from .references import References


if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin




FileNameExtractor = Callable[ [Any], PathCwd]
""" Function used to extract the source uri (as string) and the actual PathCwd of the file
    related to a decorated plugin method.
    The received arguments are those of the decorated method, _self included_.

    The function has to return a PathCwd (Path), which is the actual path of the related
    file on the disk, rooted at the cwd.
"""


DecoratorEventCbk = Callable
""" Arguments of the decorated method, prepended with the uri and the real path of the "file of
    interest" for the decorated method. Hence:

    `function( src:UriPath, real:PathCwd, plugin, *args, **kw )`

    WARNING:

        src may be relative to the docs_dir, or to the cwd, depending on the decorated function,
        and is not an actual path anyway. It's more like the identifier of a file in mkdocs.
"""



#---------------------------------------------------------



def clear_other_snippets(*_,**__):
    logger.debug("Clear snippets unrelated to the plugin")
    StaticHandler.HANDLER.references.clear_other_snippets()


def clear_references_from_src(cwd_path, *_,**__):
    logger.debug(f"Clear previous references for the file {cwd_path=!r}")
    StaticHandler.HANDLER.clear_file_related_data(cwd_path)


def mark_src_as_up_to_date(cwd_path, *_, **__):
    logger.debug(f"Mark file {cwd_path=!r} as updated")
    StaticHandler.HANDLER.files_tracker.mark_file(cwd_path)










class StaticHandler:
    """
    Global object that will manage everything needed to handle more gracefully the performances
    of the plugin:

    * keep track of the last modified files (timestamp based)
    * wrap the hooks of the plugin to bypass useless calls, depending on the plugin configuration
    * store the global internal state(s)


    One unique instance will be created during the on_config event if it doesn't already exist,
    and then reused on each new build.
    Teardown automatically occurs with on_shutdown, when the thread is closed.
    """


    used_refs: AddressesUsagePool
    """ Associate all the references to identifiers with the page they are used in:
            id_to_sources:  reference -> set of all pages using this reference
            source_to_ids:  source    -> set of all references used in that source file
    """

    references: References
    """ Associate all the identifiers with their actual target:
            id_to_data:     "--duh" -> target of the defined identifier
            id_to_source:   "--duh" -> source = the file where the identifier is declared:
                                       page.file.src_uri or filename (slash separated)
                                       (this is the target link, stripped of any anchor, actually)
            source_to_ids:  for the given file, register all the identifiers defined there
    """

    HANDLER: 'StaticHandler' = None
    ACTIVATE_CASH = True

    # testing purpose only
#    LOCK = Lock()                                                           ##>LOCK
#    USE_LOCK = False                                                         ##<LOCK


    @classmethod
    def inject(cls, plugin:'AddressAddressesPlugin', config:MkDocsConfig):
        """
        Inject the plugin in the StaticHandler "context", handling its singleton instance.
        """

        cls.ACTIVATE_CASH = plugin.config.activate_cache

#        if cls.USE_LOCK:                                                    ##>LOCK
#            if cls.LOCK.locked():
#                raise ConcurrentPluginError("Already locked")
#            cls.LOCK.acquire()                                              ##<LOCK

        # if the cache is not activated, also rebuild a fresh instance:
        if not cls.HANDLER:
            plugin.log_more("Initiate cache manager")
            cls.HANDLER = StaticHandler(plugin, config)

        elif not cls.ACTIVATE_CASH:
            plugin.log_more("Cache manager deactivated")
            cls.HANDLER.clean_up(plugin, config)

        else:
            plugin.log_more("Refresh cache manager")
            cls.HANDLER.rebuild_with(plugin, config)

        return cls.HANDLER



#    @classmethod                                                            ##>LOCK
#    def unlock(cls):
#        if not cls.LOCK.locked():
#            raise ConcurrentPluginError("Wasn't locked yet")
#        cls.LOCK.release()                                                  ##<LOCK




    #---------------------------------------------------------------------------


    _config:       MkDocsConfig                 # archive just in case (but most likely useless)
    _plugin:      'AddressAddressesPlugin'      # archive just in case
    files_tracker: FilesTracker


    def __init__(self, plugin:'AddressAddressesPlugin', config:MkDocsConfig):
        """ Instantiate only with an already validated/preprocessed plugin instance """
        self.clean_up(plugin, config)


    def clean_up(self, plugin:'AddressAddressesPlugin'=None, config:MkDocsConfig=None):
        """
        Reset the internals of the global StaticHandler, but keeping the current global innstance
        alive (testing purpose, mostly)
        """
        if plugin:
            self._plugin = plugin
        if config:
            self._config = config
        self.used_refs     = AddressesUsagePool(uni_docs_dir=plugin.uni_docs_dir)
        self.references    = References(uni_docs_dir=plugin.uni_docs_dir)
        self.files_tracker = FilesTracker()
        return self


    def rebuild_with(self, plugin:'AddressAddressesPlugin', config:MkDocsConfig):
        """ On a new build, use the fresh instances, checking of a global rebuild of the
            internal data must be made or note.
            This happens on_build only.
        """
        self._plugin = plugin
        self._config = config
        self.used_refs.archive_current(plugin.uni_docs_dir)
        self.references.archive_current(plugin.uni_docs_dir)

        # clean up from any outdated files (deleted since last build)
        uris_and_paths = self.files_tracker.get_uris_and_paths()
        cleared = []
        for cwd_src in uris_and_paths:
            if not cwd_src.is_file():
                cleared.append(cwd_src)
                self.files_tracker.remove_as_cwd_path(cwd_src)
                self.clear_file_related_data(cwd_src)

        if cleared:
            msg = 'Removed files:' + ''.join(f'\n   {file}' for file in cleared)
            self._plugin.log_more(msg)


    def clear_file_related_data(self, src:PathCwd):
        """ Remove all references or snippets data related to the given src_uri """
        logger.debug(f"Cleared references of the { src !r} file")
        self.used_refs.remove_source(src)
        self.references.remove_source(src)


    def get_missing_refs_infos(self):
        return self.used_refs.get_missing_refs_infos( self.references.id_to_data )



    #---------------------------------------------------------------------------



    @classmethod
    def __cached_call_decorator(
        cls,
        filename_extractor: FileNameExtractor,
        skip_if: Callable[[Any],str],
        *,
        before_method_call:  DecoratorEventCbk=None,
        after_method_call:   DecoratorEventCbk=None,
        # finally_after_call:  DecoratorEventCbk=None,
        # on_errors: Dict[Type[Exception], DecoratorEventCbk] = {},
        apply_if_not_called: DecoratorEventCbk=None,
    ):
        """ Decorator to add caching logic to plugin methods.

            Overall idea: Each call must be relatable to one file on the disk. So:

            1. Extract the name of the related file, using the filename_extractor callback.
               WARNING: the filename_extractor callback has to return the PathCwd of the file.

            2. Check if:
                - the plugin config ask for skipping the call or not, using a specifically named
                  method: skip_{...}_if(plugin, *args, **kw). This method can apply other kind of
                  considerations to forbid the call, if needed.
                - that file has been modified or not since the last call

            3.a. If the call to the decorated method must be done, this call is surrounded with
                 other callbacks calls, so that logic to handle things related to the cached data
                 can be applied:
                    a. Call the "before" callback
                    b. Call the decorated method
                    c. Call the "after" callback
                    d. Call the "finally after" callback
            3.b. If the method must not be called, the apply_if_not_called callback is called.

            All callbacks are optional, except the filename extractor and the skip ones.

            WARNING:
              * Exceptions in the call are NOT handled here, so the "after_method_call" logic
                might not always be applied. This is to keep in mind, but generally, this means
                the files won't be marked as up to date, so it's ok.
              * On the other hand, Exceptions that _are_ caught in the wrapped function might
                cause problems if a failure has occurred but the "after_method_call" is executed
                anyway. Generally, it's not a big problem, because the user will have the feedback
                in the console, will update the file, and so everything will be alright. But to
                limit the potential problems, better to wrap this decorator with the auto_logger
                one than the opposite (since the logger might swallow some exceptions...).
        """
        # pylint: disable=multiple-statements, protected-access

        def decorator(method:Callable):

            # Note: The checks on ACTIVATED_CASH and the cbks definitions must be made at run time
            #       only, incase the state changes in between calls.

            @wraps(method)
            def wrapper(self:'AddressAddressesPlugin', *args, **kwargs):

                if self.plugin_off:
                    return

                if not cls.HANDLER:
                    raise AddresserError(
                        "Wrong use of one of the StaticHandler decorators: the global HANDLER "
                        "instance should have been already defined"
                    )

                cached_file = filename_extractor(self, *args, **kwargs)
                # cls.HANDLER.__check_src_uri(cached_file)
                to_skip     = skip_if(self, *args, **kwargs)
                up_to_date  = to_skip or cls.HANDLER.files_tracker.is_file_up_to_date(cached_file)

                need_to_run = not (to_skip or up_to_date)
                if not cls.ACTIVATE_CASH or need_to_run:
                    # Always use this branch if caching is deactivated

                    if before_method_call and cls.ACTIVATE_CASH:
                        before_method_call(cached_file, self, *args, **kwargs)

                    # If the cash isn't activated, the skip check did not occur so, done here
                    # so that the behaviors stay consistent:
                    if cls.ACTIVATE_CASH or not skip_if(self, *args, **kwargs):
                        method(self, *args,  **kwargs)

                    if after_method_call and cls.ACTIVATE_CASH:
                        after_method_call(cached_file, self, *args, **kwargs)

                elif apply_if_not_called:
                    apply_if_not_called(cached_file, self, *args, **kwargs)

            return wrapper
        return decorator





    #---------------------------------------------------------------------------
    #       Decorators (as class methods) to wrap around the plugin logic
    #---------------------------------------------------------------------------



    @classmethod
    def snippets_extraction_cached_with(cls, filename_extractor:Callable, skip_if:Callable):
        """ If the snippets file is valid and has been modified (or is not cached yet):
                - before: clear any previous data for snippets entries that wouldn't be related
                          to the plugin itself.
                - call:   apply the logic of extraction
                - after:  register the last modification date of the file
        """
        return cls.__cached_call_decorator(
            filename_extractor,  skip_if,
            before_method_call = clear_other_snippets,
            after_method_call  = mark_src_as_up_to_date,
        )



    @classmethod
    def external_links_cached_with(cls, filename_extractor:Callable, skip_if:Callable):
        """ TODO: finish (or just check it's ok like this...) """

        return cls.__cached_call_decorator(
            filename_extractor,  skip_if,
            before_method_call = clear_references_from_src,
            after_method_call  = mark_src_as_up_to_date,
        )



    @classmethod
    def other_files_cached_with(cls, filename_extractor:Callable, skip_if:Callable):
        """ TODO: finish (or just check it's ok like this...) """

        return cls.__cached_call_decorator(
            filename_extractor,  skip_if,
            before_method_call = clear_references_from_src,
            after_method_call  = mark_src_as_up_to_date,
        )



    @classmethod
    def on_page_content_cached_with(cls, filename_extractor:Callable, skip_if:Callable):
        """ 'on_page_content' gathers the information needed for the identifiers and the snippets
            only, and does so working on one single page (ie .md) file, so:

            If the file has been changed or is not cached yet:
                1. before: remove any snippet/ref related to that file
                2. call

            Note: do NOT mark the file now, otherwise on_page_context won't apply.
        """
        return cls.__cached_call_decorator(
            filename_extractor,  skip_if,
            before_method_call = clear_references_from_src,
        )



    @classmethod
    def on_page_context_cached_with(
        cls,
        filename_extractor:Callable,
        skip_if:Callable,
        page_arg_extractor_page_hooks:Callable,
    ):
        """ 'on_page_context' parses the rendered html file, replaces all the identifiers,
            computing the needed relative paths to reach the targets from the current file, and
            check that all the addresses in the page are compliant with the plugin rules, so:

            If the file has been changed or is not cached yet, and if the plugin is not in
            "dump_only" mode:
                1. call
                2. after: cache the page.content value and register the last modification
                          date of the file.
            If the above condition wasn't met, the cached value of page.content is restored,
            so that the built page is still the expected one.
        """

        def after_mark_page_and_cache_content(cwd_path, plugin, *args, **kw):
            cls.HANDLER.files_tracker.mark_file(cwd_path)
            page:Page = page_arg_extractor_page_hooks(plugin, *args, **kw)
            cache[cwd_path] = page.content

        def apply_restore_from_cache(cwd_path, plugin:'AddressAddressesPlugin', *args, **kw):
            plugin.log_more(f"Restore cached content for { cwd_path }")
            if cwd_path in cache:
                page:Page = page_arg_extractor_page_hooks(plugin, *args, **kw)
                page.content = cache[cwd_path]

        cache = {}     # Will never get purged until the server is stopped, but hey... ;p

        return cls.__cached_call_decorator(
            filename_extractor,  skip_if,
            after_method_call   = after_mark_page_and_cache_content,
            apply_if_not_called = apply_restore_from_cache,
        )



"""
Group of cached functions, to not recompute again and again the very same paths conversions...
"""


# pylint: disable=all

import os
from pathlib import Path
from itertools import chain

from typing import TYPE_CHECKING, Union
from functools import lru_cache

from mkdocs_addresses.static_handler.types_aliases import PathCwd



if TYPE_CHECKING:
    from .addresses_plugin import AddressAddressesPlugin





@lru_cache(None)
def to_uri(path:Union[Path,str], *segments:str):
    """ Take a string, potentially os dependent path, and rebuild a slash separated
        version from it.
        If additional segments are given, they'll be considered new directories, up
        to a final segment that will be considered a file.

        Behavior for trailing separators on @path:
            - If @segments is not given, trailing separators on @path are kept (to
              keep behaviors consistent with "hidden" index.html files in the address).
            - If any @segments is given, trailing separators on @path are removed
              before joining @path and @segments.
    """
    if isinstance(path,str):
        path = Path(path)
    joined = "/".join(path.parts + segments)
    if path.root=='/':
        joined = joined[1:]
    return joined


@lru_cache(None)
def to_os(uri:str, *segments:str):
    """ Takes a path as string and converts it to an OS dependant version """
    return Path(uri, *segments)


#---------------------------------------------------------------------


@lru_cache(None)
def get_cwd_path_from_docs_dir_and_uri(uni_docs_dir:Path, src_uri:str) -> PathCwd:
    """ Specific to page_filename_arg_extractor... """
    return uni_docs_dir / src_uri



from argparse import Namespace
import os
from pathlib import Path
import re

from typing import Any, Callable, List, Literal, Optional, Tuple, TYPE_CHECKING

from mkdocs.config import config_options  as C
from mkdocs.config.base import Config
from mkdocs.config.defaults import MkDocsConfig

from mkdocs_addresses import path_manager

from .exceptions import AddresserConfigError, InvalidOperationError

if TYPE_CHECKING:
    from .addresses_plugin import AddressAddressesPlugin






# Definition used for mkdocstrings
class PluginOptions(Namespace):


    activate_cache: bool = True
    """
    Exploring and cross-checking all references in the documents can be rather time consuming
    because of the way it is done in the plugin (exploring the resulting html content) which
    means the plugin could significantly slow down the serve operations when the number of
    documents or their size becomes rather big.

    To limit this problem, the plugin uses a cache (activated by default).

    If activated, only the recently modified markdown files will be explored again on next
    serve. This allows to drastically reduce the serving time, while the plugin still can
    check for outdated references in the unchanged pages, since their content is cached.

    Note that when mkdocs starts, all pages must be explored at least once, so the first serve
    operation can be very slow on big documentations.
    """


    dump_action: str = 'normal'
    """
    The plugin normally dumps information about all the available references (see
    [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
    option) on each serve.

    The `dump_action` determines what the plugin does about the
    [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file].

    * `normal`: the plugin recreates the informations on each serve.
    * `none`: the plugin uses its normal behavior, but the
      [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
      is not updated.
    * `only`: the plugin will skip all the addresses verifications in the documentation files and
      will only gather the references definitions, drop all of them in the
      [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file],
      then raise `DumpOnlyError` to stop serving.

        This option may come in handy if you need to regenerate the code snippets while something
        normally causes a crash during the addresses verifications.
    """


    dump_file_vsc: str = '.vscode/links.code-snippets'
    dump_file_txt: str = 'addresses_identifiers.txt'
    dump_snippets_file: str = ''
    """
    Define the location of the file where the information about all the references will be
    saved (relative to the cwd).

    There are a number of restrictions on the possible values of this option:

    * The targeted file cannot be in the `docs_dir` (it should never be anyway).
    * The extension of the file must be compatible with the value chosen for the
    [`use_vsc`][mkdocs_addresses.config_plugin.PluginOptions.use_vsc]
    option.
    * If left undefined, the filename is chosen depending on the
    [`use_vsc`][mkdocs_addresses.config_plugin.PluginOptions.use_vsc] option.

    | `use_vsc` | Default `dump_snippets_file` | File extension |
    |:-|:-|:-|
    | `true` | `.vscode/links.code-snippets` | `.code-snippets` |
    | `false` | `addresses_identifiers.txt` | `.txt` |
    """


    external_links_file: str = ''
    """
    Location of the file, if any, where are defined the global links, like:

    ```markdown
    [mkdocs]: https://www.mkdocs.org/ "mkdocs - home"
    [id]: address "tooltip"
    ...
    ```

    If this option is defined:

    - This file cannot be in the docs_dir itself (it shouldn't anyway).
    - All links defined in this file must be absolute or the plugin will raise an error. This
      restriction can be bypassed by "white listing" the address (see
      [`links_white_list`][mkdocs_addresses.config_plugin.PluginOptions.links_white_list] and
      [`links_white_list_pattern`][mkdocs_addresses.config_plugin.PluginOptions.links_white_list_pattern]).
    - This file should be appended automatically to all files through the `pymdownx.snippets`
      markdown extension, directly in the `mkdocs.yml` file. For example:

        ```yaml
        markdown_extensions:
            - pymdownx.snippets:
                check_paths: true
                auto_append: ["docs_logistic/external-links.md"]

        plugins:
            - mkdocs-addresses:
                extenral_links_file: docs_logistic/external-links.md
        ```
    """


    fail_fast: bool = False
    """
    Defines if the errors raised by the plugin will interrupt the serve operations or will only
    result in feedback in the console.

    When writing the docs, it may be useful to pass this option to false: this way, there is no
    need to restart the serving each time a link is wrong and you can keep working, fixing the
    errors step by step.

    Note that if `fail_fast` is false, a summary of all the wrong addresses found will be
    displayed once all the pages have been checked. This is especially useful if there are
    a lot of errors and the console becomes somewhat cluttered with messages.
    """


    imgs_extension_pattern: str = "\\.(jpg|jpeg|bmp|png|svg)$"
    """
    Regex string used to identify picture files (using
    [`re.search`](https://docs.python.org/3/library/re.html#re.search){: target=_blank }).
    """


    ignore_auto_headers: bool = True
    """
    By default, mkdocs turns the content of each header into an html id that can be used as anchor.

    Considering how the plugin builds the references, and that it forbids duplicate identifiers,
    this would mean that duplicated titles would be impossible to use across the same project
    documentation. Which would be... rather unfortunate.

    There are several ways to circumvent this:

    1. Define an id attribute for the problematic headers: `## duplicate title {: #unique-id }`
      (reminder: this supposes you are using the `attr_list` markdown extension).
    1. Add the automatically generated id to the list of [`ignored_ids`][mkdocs_addresses.config_plugin.PluginOptions.ignored_ids].
    1. "White list" the id, [one way][mkdocs_addresses.config_plugin.PluginOptions.links_white_list]
      or [another][mkdocs_addresses.config_plugin.PluginOptions.links_white_list_pattern].
    1. Set `ignore_auto_headers` to true. The plugin will then attempt to spot the automatic
    identifiers and ignore them.

        Note that false negative can occur from time to time: when the plugin is searching for
        references in the html code, it tries to compare the content of the headers with the
        html id and ignores it if it is considered equivalent to the content (roughly: it
        converts the text content to a "simplified kebab-case", removing some special characters,
        then compares the result to the id).

    For example:

    * With `ignore_auto_headers` set to false:

    ```markdown
    ## Content
    ...

    ### Content

    This second title raises DuplicateIdentifierError (even with a different header level)

    ## Content {: #precedence }
    This title wouldn't cause problems, because its id will be used instead
    ```

    * With `ignore_auto_headers` set to true:

    ```markdown
    ## Content
    ... is ignored

    ## Content
    ... is also ignored

    ## Content {: #hey-there }
    ... is gathered as --hey-there

    ### Hall of fame! {: #hall-of-fame }
    This will be a false negative: the id is considered equivalent to the content,
    and this id will be ignored.
    ```
    """


    ignored_ids: List[str] = []
    """
    If defined, each string will be an html id that will be ignored when gathering and checking
    addresses. Example of definition:

    ```yaml
    plugins:
        - mkdocs-addresses:
            ignored_ids:            # or: ignored_ids: ["subtitle", "skip-this-id"]
                - subtitle
                - skip-this-id
    ```

    Note:
        Ids starting with double underscores will always be ignored (because some plugins use
        them for their own purpose).
    """


    ignored_classes: List[str] = []
    """
    If defined, each string will be an html class that will be ignored when gathering and
    checking addresses. Example of definitions:

    ```yaml
    plugins:
        - mkdocs-addresses:
            ignored_classes:        # or: ignored_classes: ["class-name-1", "non-gathered"]
                - class-name-1
                - non-gathered
    ```

    Note:
        The class "headerlink" will always be skipped (this allows to ignore the permalinks
        generated by the `toc` markdown extension).
    """


    inclusions: List[str] = []
    """
    Slash separated paths, relative to the cwd, that must point to directories containing the
    files you may use for "inclusions" with the `pymdownx.snippets` markdown extension, using:
    `--<8--`.

    If defined, each directory will be added to the list of watched directories, and their
    content will be recursively explored: the plugin will build the code snippets to
    include those files in the docs pages. See [TODO](TODO).

    Example of included snippet:

    `--8<-- "path_from_cwd_to_included_file.with_extension"`

    Note:
        These files are not supposed to be in the `docs_dir` folder in the first place. If they
        are, mkdocs will complain about files not included in the nav anyway.
    """


    links_white_list: List[str] = []
    """
    List of links and/or ids that will ignored when gathering the references, and also when
    checking the addresses. For example:

    ```yaml
    plugins:
        - mkdocs-addresses:
            links_white_list:        # or   links_white_list: ["/media/toto/local_file", "toto"]
                - /media/toto/local_file
                - toto
    ```
    """


    links_white_list_pattern: str = "https?://|ftps?://|file:///|www\\."
    """
    Regex string: all addresses or identifiers that will
    [`re.match`](https://docs.python.org/3/library/re.html#re.match){: target=_blank } with it will be ignored.

    This allows to use absolute addresses, external links, or any other weird case you could
    imagine, so that the plugin doesn't raise an error for them.
    """


    more_verbose: bool = False
    """
    Controls the verbosity level of the plugin, increasing it slightly, without increasing the
    verbosity level of mkdocs logger itself.

    If mkdocs is run in verbose mode, this option is ignored and the verbose mode also applies
    to the plugin.
    """


    plugin_off: bool = False
    """
    If true, completely deactivate the plugin. This may allow to speed up the serve steps if needed.
    """


    use_vsc: bool = True
    """
    Controls the kind of file that will be dumped by the plugin.

    * If true, a VSC `.code-snippets` file will be created at the end of the `on_post_build` event,
    at the location defined by the [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
    option.
    * If false, the plugin still works the same, but you will have to work without the code snippets.
    So the plugin generates a `.txt` file containing all the gathered references and showing what
    the target for each of them.

    Note:
        Keep in mind that this option determines the kind of file used for the
        [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
        option, meaning that, if you assigned a filename for this option, its extension has to
        match with the value of the `use_vsc` option.

        See the section above, explaining the
        [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
        option.
    """


    verify_only: List[str] = []
    """
    If items are defined, the correctness of the addresses will be checked only in those
    markdown files. All other operations are done normally.

    The location must slash separated path, relative to the docs_dir (and _not_ to the cwd).
    For example, to only check "docs/subdir/blabla.md, use:

    ```yaml
    plugins:
        - mkdocs-addresses:
            verify_only:
                - subdir/blabla.md
    ```
    """








PLUGIN_MARK = "VSC-REF"

DUMP_OPTIONS = tuple('normal only none'.split())
NORMAL, DUMP_ONLY, NO_DUMP = DUMP_OPTIONS
assert PluginOptions.dump_action in DUMP_OPTIONS, "Invalid plugin setup"




# Config types definitions...:

ConfigErrors   = List[Tuple[str,bool,str,Optional[Exception]]]
ConfigWarnings = List[Tuple[str,str]]
FatalCbk       = Callable





class AddressAddressesConfig(Config):

    activate_cache = C.Type(bool, default=PluginOptions.activate_cache)

    dump_action = C.Choice(DUMP_OPTIONS, default=PluginOptions.dump_action)

    dump_snippets_file = C.Type(str, default='')      # Automatically overridden at validation time

    external_links_file = C.Type(str, default=PluginOptions.external_links_file)

    fail_fast = C.Type(bool, default=PluginOptions.fail_fast)

    imgs_extension_pattern = C.Type(str, default=PluginOptions.imgs_extension_pattern)

    ignore_auto_headers = C.Type(bool, default=PluginOptions.ignore_auto_headers)

    ignored_ids = C.ListOfItems(C.Type(str), default=PluginOptions.ignored_ids)

    ignored_classes = C.ListOfItems(C.Type(str), default=PluginOptions.ignored_classes)

    inclusions = C.ListOfItems(C.Type(str), default=PluginOptions.inclusions)

    links_white_list = C.ListOfItems(C.Type(str), default=PluginOptions.links_white_list)

    links_white_list_pattern = C.Type(str, default=PluginOptions.links_white_list_pattern)

    more_verbose = C.Type(bool, default=PluginOptions.more_verbose)

    plugin_off = C.Type(bool, default=PluginOptions.plugin_off)

    use_vsc = C.Type(bool, default=PluginOptions.use_vsc)

    verify_only = C.ListOfItems(C.File(exists=True), default=PluginOptions.verify_only)



    #----------------------------------------------



    def validate_and_process(self, config:MkDocsConfig):
        """ Performs the basic, then extras validations on each property, and also apply the
            post conversions to the data structures that need it.
            Add on the fly the needed fields from the mkdocs config.
        """

        # Pre-conversions:
        #-----------------

        if not self.dump_snippets_file:
            self.dump_snippets_file = PluginOptions.dump_file_vsc if self.use_vsc else PluginOptions.dump_file_txt


        # Perform "post" conversions:       (yeah, "post"... lol...)
        #----------------------------

        maybe_abs = path_manager.to_os(config['docs_dir'])
        self.uni_docs_dir = maybe_abs.absolute().relative_to(Path.cwd())

        for prop in 'docs_dir  use_directory_urls'.split():
            setattr(self, prop, config[prop])


        post_conversions: Tuple[Callable[[str],Any], str] = [
            (re.compile, 'imgs_extension_pattern  links_white_list_pattern'),
            (set,        'ignored_ids  ignored_classes  links_white_list  verify_only  inclusions'),
        ]
        for conversion,props in post_conversions:
            for prop in props.split():
                value = getattr(self, prop)
                try:
                    post = conversion(value)
                    setattr(self, prop, post)
                except Exception as e:                              # pylint: disable=all
                    e.args = ( f"{ prop }: { e.args[0] }",)
                    raise

        self.ignored_classes.add('headerlink')      # this avoids to consider mkdocs permalinks in headers, when used

        # More specific validations:
        #---------------------------

        self.__validate_inclusions_dir()
        self.__validate_links_white_list_pattern()
        self.__validate_external_links_file()
        self.__validate_dump_snippets_file()
        self.__validate_verify_only_md_files()
        self.__validate_vsc_and_dump_filename_consistency()



    #----------------------------------------------

    def __check_not_in_docs(self, prop:str, path:str):
        rel = os.path.relpath(path, self.uni_docs_dir)
        if not rel.startswith('..'):
            raise AddresserConfigError(
                f"{prop}: {path!r} should not be inside the docs_dir directory"
            )


    def __validate_inclusions_dir(self):
        self.inclusions = sorted(map( path_manager.to_os, self.inclusions))
        oops = [location for location in self.inclusions if not location.is_dir() ]
        if oops:
            raise AddresserConfigError(
                f"'inclusions' should be a list of directories, but found invalid paths: ({ oops !r})"
            )

        for path in self.inclusions:
            self.__check_not_in_docs('inclusions', path)


    def __validate_links_white_list_pattern(self):
        if not self.links_white_list_pattern.pattern:      # this is now a pattern, not a string anymore!
            raise AddresserConfigError(
                "links_white_list_pattern should never be an empty string. Note that not using at "
                "least the default pattern will definitely cause problems..."
            )


    def __validate_external_links_file(self):
        if not self.external_links_file:
            return

        self.external_links_file = p = path_manager.to_os(self.external_links_file)
        if not p.is_file() or not p.suffix=='.md':
            raise AddresserConfigError(
                "'external_links_file' is configured, but is not an existing markdown file "
                f'("{ self.external_links_file }")'
            )
        self.__check_not_in_docs('external_links_file', self.external_links_file)


    def __validate_dump_snippets_file(self):
        self.dump_snippets_file = path_manager.to_os(self.dump_snippets_file)
        self.__check_not_in_docs('dump_snippets_file', self.dump_snippets_file)


    def __validate_vsc_and_dump_filename_consistency(self):
        snip_file:Path   = self.dump_snippets_file
        vsc_and_snippets = self.use_vsc and snip_file.suffix=='.code-snippets'
        not_vsc_and_txt  = not self.use_vsc and snip_file.suffix=='.txt'
        if not (vsc_and_snippets or not_vsc_and_txt):
            raise AddresserConfigError(
                "Inconsistent configuration:\n"
                f"   use_vsc: {self.use_vsc}\n"
                f"   dump_snippets_file: {self.dump_snippets_file}"
            )


    def __validate_verify_only_md_files(self):
        invalids = [ file for file in self.verify_only
                          if not path_manager.to_os(file).suffix=='.md']
        if invalids:
            raise AddresserConfigError(
                "All options for `verify_only` should be valid markdown files, relative to the cwd.\n"
                +"Invalid entries:" + "".join('\n    '+file for file in invalids)
            )





class ConfigDescriptor:
    """ Dynamic relay data descriptor, to articulate the plugin and its configuration """

    prop: str

    def __set_name__(self, _, prop:str):
        self.prop = prop

    def __get__(self, obj:'AddressAddressesPlugin', _:type):
        return getattr(obj.config, self.prop)

    def __set__(self, obj:'AddressAddressesPlugin', value:Any):
        # setter needed to make this a data descriptor, but should never be used.
        raise InvalidOperationError(
            "Config properties should never be reassigned from the plugin instance."
        )
